# Generated by Django 3.1.2 on 2020-10-19 07:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('units', '0002_auto_20201019_0718'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customunit',
            name='symbol',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='Symbol'),
        ),
    ]



# Generated by Django 3.0.5 on 2020-07-22 07:52

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Rate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False,
                                        verbose_name='ID')),
                ('key', models.CharField(db_index=True, default=None, max_length=255, null=True)),
                ('value_date', models.DateField()),
                ('value', models.FloatField(default=0)),
                ('currency', models.CharField(max_length=3)),
                ('base_currency', models.CharField(default='EUR', max_length=3)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT,
                                           related_name='rates', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-value_date'],
            },
        ),
        migrations.AddIndex(
            model_name='rate',
            index=models.Index(fields=['currency', 'base_currency', 'value_date'],
                               name='rates_rate_currenc_6a7540_idx'),
        ),
        migrations.AddIndex(
            model_name='rate',
            index=models.Index(fields=['key', 'currency', 'base_currency', 'value_date'],
                               name='rates_rate_key_8c7b0f_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='rate',
            unique_together={('key', 'currency', 'base_currency', 'value_date')},
        ),
    ]



"""
Serializers for Currencies module
"""
from rest_framework import serializers

from .models import Currency


class CurrencySerializer(serializers.Serializer):
    """
    Serializer for Currency class
    """
    code = serializers.CharField(label="ISO-4217 code of the currency", read_only=True)
    name = serializers.CharField(label="Human readable name of the currency", read_only=True)
    currency_name = serializers.CharField(label="name of the currency", read_only=True)
    exponent = serializers.IntegerField(label="ISO-4217 exponent", read_only=True)
    number = serializers.IntegerField(label="ISO-4217 numeric identifier", read_only=True)
    value = serializers.CharField(read_only=True)
    symbol = serializers.SerializerMethodField(label="Symbol of the currency")

    @staticmethod
    def get_symbol(obj: Currency):
        """
        Get symbol from currency
        """
        return obj.symbol

    def create(self, validated_data: dict) -> Currency:
        """
        Returns a Currency object from cleaned data
        """
        return Currency(code=validated_data.get('code'))

    def update(self, instance: Currency, validated_data: dict) -> Currency:
        """
        No update
        """
        return instance



"""
Viewsets for currencies module APIs
"""

import logging
import statistics
from datetime import date, timedelta

from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from geocurrency.countries.serializers import CountrySerializer
from geocurrency.rates.serializers import RateSerializer
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import ReadOnlyModelViewSet

from .models import Currency, CurrencyNotFoundError
from .serializers import CurrencySerializer


class CurrencyViewset(ReadOnlyModelViewSet):
    """
    View for currency
    """
    lookup_field = 'code'

    currencies_response = openapi.Response('List of currencies', CurrencySerializer)
    currency_response = openapi.Response('Currency detail', CurrencySerializer)
    ordering = openapi.Parameter('ordering', openapi.IN_QUERY,
                                 description="Sort on code, name, currency_name, "
                                             "exponent, number, value. "
                                             "Prefix with - for descending sort",
                                 type=openapi.TYPE_STRING)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[ordering, ], responses={200: currencies_response})
    def list(self, request, *args, **kwargs):
        """
        List of currencies
        """
        currencies = Currency.all_currencies(ordering=request.GET.get('ordering', 'name'))
        serializer = CurrencySerializer(currencies, many=True, context={'request': request})
        return Response(serializer.data)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(responses={200: currency_response})
    def retrieve(self, request, code, *args, **kwargs) -> Response:
        """
        Retrieve single record based on iso4217 code
        """
        try:
            currency = Currency(code)
            serializer = CurrencySerializer(currency, context={'request': request})
            return Response(serializer.data)
        except CurrencyNotFoundError:
            return Response('Currency not found', status=status.HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(method='get', responses={200: CountrySerializer})
    @action(['GET'], detail=True, url_path='countries', url_name="get_countries")
    def get_countries(self, request, code) -> Response:
        """
        Get all countries for the currency
        :param request: HTTP request
        :param code: Currency id
        :return: List of countries
        """
        try:
            currency = Currency(code)
            serializer = CountrySerializer(currency.countries,
                                           many=True,
                                           context={'request': request})
            return Response(serializer.data)
        except CurrencyNotFoundError as e:
            logging.error(e)
            return Response('Currency not found', status=status.HTTP_404_NOT_FOUND)

    from_date = openapi.Parameter('from_date', openapi.IN_QUERY,
                                  description="From date (YYYY-MM-DD)",
                                  type=openapi.TYPE_STRING)
    to_date = openapi.Parameter('to_date', openapi.IN_QUERY, description="To date (YYYY-MM-DD)",
                                type=openapi.TYPE_STRING)
    base_currency = openapi.Parameter('base', openapi.IN_QUERY,
                                      description="Base currency (defaults to EUR)",
                                      type=openapi.TYPE_STRING)
    currency = openapi.Parameter('base_currency', openapi.IN_QUERY, description="currency",
                                 type=openapi.TYPE_STRING)
    key = openapi.Parameter('key', openapi.IN_QUERY, description="custom key",
                            type=openapi.TYPE_STRING)

    @swagger_auto_schema(method='get',
                         manual_parameters=[from_date, to_date, base_currency, key],
                         responses={200: RateSerializer})
    @action(['GET'], detail=True, url_path='rates', url_name="get_currency_rates")
    def get_rates(self, request, code: str,
                  key: str = None, base_currency: str = 'EUR',
                  from_date: date = date.today(), to_date: date = date.today()) -> Response:
        """
        Get conversion rates for the currency
        :param request: HTTP request
        :param code: Currency id
        :param key: optional custom key
        :param base_currency: optional base currency, defaults to EUR
        :param from_date: optional limit results to rates range in time
        :param to_date: optional limit results to rates range in time
        :return: List of rates
        """
        try:
            c = Currency(code)
            user = None
            if request.user and request.user.is_authenticated:
                user = request.user
            rates = c.get_rates(user=user, key=key, base_currency=base_currency,
                                start_date=from_date, end_date=to_date)
            serializer = RateSerializer(rates, many=True, context={'request': request})
            return Response(serializer.data)
        except CurrencyNotFoundError:
            return Response('Currency not found', status=status.HTTP_404_NOT_FOUND)

    @staticmethod
    def _get_rates(
            base_currency: str,
            target_currency: str,
            from_date: date,
            to_date: date) -> dict:
        """
        Generates list of rates and stats
        :param base_currency: rate to currency
        :param target_currency: rate from currency
        :param from_date: lower date range
        :param to_date: higher date range
        """
        rates = []
        dates = []
        for i in range((to_date - from_date).days + 1):
            d = from_date + timedelta(i)
            cd, reference, rate = base_currency.convert(
                target_currency=target_currency,
                conversion_date=d
            )
            rates.append(rate)
            dates.append(
                {
                    'conversion_date': cd.strftime('%Y-%m-%d'),
                    'currency': base_currency,
                    'reference': target_currency,
                    'rate': rate
                }
            )
        stat = {}
        if len(rates) > 1:
            stat = {
                'avg': statistics.mean(rates),
                'median': statistics.median(rates),
                'max': max(rates),
                'min': min(rates),
                'std_deviation': statistics.stdev(rates)
            }
        return {'rates': dates, 'statistics': stat}



"""
Units module APIs viewsets
"""

import json
import logging

from django.db import models
from django.http import HttpResponseForbidden, HttpRequest
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from django_filters import rest_framework as filters
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from geocurrency.converters.models import ConverterLoadError
from geocurrency.converters.serializers import ConverterResultSerializer
from geocurrency.core.helpers import validate_language
from geocurrency.core.pagination import PageNumberPagination
from rest_framework import permissions, status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_400_BAD_REQUEST
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet, ModelViewSet

from . import DIMENSIONS
from .exceptions import UnitConverterInitError, UnitSystemNotFound, UnitNotFound, \
    DimensionNotFound, ExpressionCalculatorInitError, UnitValueError
from .filters import CustomUnitFilter
from .forms import CustomUnitForm
from .models import UnitSystem, UnitConverter, Dimension, CustomUnit, ExpressionCalculator
from .permissions import CustomUnitObjectPermission
from .renderers import UnitsRenderer
from .serializers import UnitSerializer, UnitSystemSerializer, \
    UnitConversionPayloadSerializer, DimensionSerializer, \
    DimensionWithUnitsSerializer, CustomUnitSerializer, ExpressionSerializer, \
    CalculationPayloadSerializer, CalculationResultSerializer


class UnitSystemViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'system_name'

    language_header = openapi.Parameter('Accept-Language', openapi.IN_HEADER,
                                        description="language",
                                        type=openapi.TYPE_STRING)
    language = openapi.Parameter('language', openapi.IN_QUERY,
                                 description="language",
                                 type=openapi.TYPE_STRING)

    unit_systems_response = openapi.Response('List of unit systems', UnitSystemSerializer)
    unit_system_response = openapi.Response('Dimensions and units in a system',
                                            UnitSystemSerializer)

    ordering = openapi.Parameter('ordering', openapi.IN_QUERY,
                                 description="Sort on name "
                                             "Prefix with - for descending sort",
                                 type=openapi.TYPE_STRING)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header, ordering],
                         responses={200: unit_systems_response})
    def list(self, request):
        """
        List UnitSystems
        """
        ordering = request.GET.get('ordering', 'system_name')
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['system_name']:
            ordering = 'system_name'
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        us = UnitSystem(fmt_locale=language)
        us = [{'system_name': s} for s in sorted(us.available_systems(), reverse=descending)]
        return Response(us, content_type="application/json")

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header],
                         responses={200: unit_system_response})
    def retrieve(self, request, system_name):
        """
        Retrieve UnitSystem
        """
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            serializer = UnitSystemSerializer(us, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except UnitSystemNotFound as e:
            return Response("Unknown unit system: " + str(e), status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header, ordering],
                         responses={200: DimensionSerializer})
    @action(methods=['GET'], detail=True, name='dimensions', url_path='dimensions')
    def dimensions(self, request, system_name):
        """
        List Dimensions of a UnitSystem
        """
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        ordering = request.GET.get('ordering', 'name')
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            serializer = DimensionSerializer(us.available_dimensions(ordering=ordering),
                                             many=True,
                                             context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except UnitSystemNotFound as e:
            return Response("Unknown unit system: " + str(e), status=HTTP_404_NOT_FOUND)


class UnitViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'unit_name'
    language_header = openapi.Parameter('Accept-Language', openapi.IN_HEADER,
                                        description="language",
                                        type=openapi.TYPE_STRING)
    language = openapi.Parameter('language', openapi.IN_QUERY, description="language",
                                 type=openapi.TYPE_STRING)
    dimension = openapi.Parameter('dimension', openapi.IN_QUERY, description="Unit dimension",
                                  type=openapi.TYPE_STRING)
    key = openapi.Parameter('key', openapi.IN_QUERY,
                            description="key",
                            type=openapi.TYPE_STRING)
    units_response = openapi.Response('List of units in a system', UnitSerializer)
    unit_response = openapi.Response('Detail of a unit', UnitSerializer)
    dimension_response = openapi.Response('List of units per dimension',
                                          DimensionWithUnitsSerializer)
    ordering = openapi.Parameter('ordering', openapi.IN_QUERY,
                                 description="Sort on fields name or code. "
                                             "Prefix with - for descending sort",
                                 type=openapi.TYPE_STRING)

    @swagger_auto_schema(manual_parameters=[dimension, key,
                                            ordering,
                                            language, language_header],
                         responses={200: units_response})
    def list(self, request: HttpRequest, system_name: str):
        """
        List Units, ordered, filtered by key or dimension, translated in language
        """
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        ordering = request.GET.get('ordering', 'name')
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['code', 'name']:
            ordering = 'name'
        try:
            key = request.GET.get('key', None)
            user = request.user if hasattr(request,
                                           'user') and request.user.is_authenticated else None
            us = UnitSystem(system_name=system_name, fmt_locale=language, user=user, key=key)
            units = []
            dimension_param = request.GET.get(key='dimension')
            if dimension_param:
                try:
                    dimension = Dimension(unit_system=us, code=dimension_param)
                    units = dimension.units(user=user, key=key)
                except DimensionNotFound:
                    return Response('Invalid dimension filter', status=status.HTTP_400_BAD_REQUEST)
            else:
                available_units = us.available_unit_names()
                if available_units:
                    units = [us.unit(unit_name=unit_name) for unit_name in available_units]
            units = sorted(units, key=lambda x: getattr(x, ordering), reverse=descending)
            serializer = UnitSerializer(units, many=True, context={'request': request})
            return Response(serializer.data)
        except UnitSystemNotFound:
            return Response('Invalid Unit System', status=status.HTTP_404_NOT_FOUND)

    @swagger_auto_schema(manual_parameters=[key, language, language_header],
                         responses={200: dimension_response})
    @action(['GET'], detail=False, name='units per dimension', url_path='per_dimension')
    def list_per_dimension(self, request: HttpRequest, system_name: str):
        """
        List Units grouped by dimension, filter on key, translated
        """
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        try:
            key = request.GET.get('key', None)
            user = request.user if hasattr(request,
                                           'user') and request.user.is_authenticated else None
            us = UnitSystem(system_name=system_name, fmt_locale=language, user=user, key=key)
            dimensions = [Dimension(unit_system=us, code=code) for code in DIMENSIONS.keys()]
            serializer = DimensionWithUnitsSerializer(dimensions, many=True,
                                                      context={'request': request})
            return Response(serializer.data)
        except UnitSystemNotFound as e:
            logging.warning(str(e))
            return Response('Invalid Unit System', status=status.HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[key, language, language_header],
                         responses={200: unit_response})
    def retrieve(self, request: HttpRequest, system_name: str, unit_name: str):
        """
        Get unit information for unit in unit system
        """
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        try:
            key = request.GET.get('key', None)
            user = request.user if hasattr(request,
                                           'user') and request.user.is_authenticated else None
            us = UnitSystem(system_name=system_name, fmt_locale=language, user=user, key=key)
            unit = us.unit(unit_name=unit_name)
            serializer = UnitSerializer(unit, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (UnitSystemNotFound, UnitNotFound):
            return Response("Unknown unit", status=HTTP_404_NOT_FOUND)

    @swagger_auto_schema(manual_parameters=[language, language_header, ordering],
                         responses={200: units_response})
    @action(methods=['GET'], detail=True, url_path='compatible', url_name='compatible_units')
    def compatible_units(self, request: HttpRequest, system_name: str, unit_name: str):
        """
        List compatible Units
        """
        language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
        ordering = request.GET.get('ordering', 'name')
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['code', 'name']:
            ordering = 'name'
        try:
            key = request.GET.get('key', None)
            user = request.user if hasattr(request,
                                           'user') and request.user.is_authenticated else None
            us = UnitSystem(system_name=system_name, fmt_locale=language, user=user, key=key)
            unit = us.unit(unit_name=unit_name)
            compatible_units = sorted([us.unit(unit_name=cunit) for cunit in
                                       map(str, unit.unit.compatible_units())],
                                      key=lambda x: getattr(x, ordering),
                                      reverse=descending)
            serializer = UnitSerializer(compatible_units, many=True, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (UnitSystemNotFound, UnitNotFound):
            return Response("Unknown unit", status=HTTP_404_NOT_FOUND)


class ConvertView(APIView):
    """
    Convert Units API
    """

    @swagger_auto_schema(request_body=UnitConversionPayloadSerializer,
                         responses={200: ConverterResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency and date to a reference currency
        :param request: HTTP request
        """
        cps = UnitConversionPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST,
                            content_type="application/json")
        cp = cps.create(cps.validated_data)
        user = None
        if request.user and request.user.is_authenticated:
            user = request.user
        key = request.POST.get('key', None)
        try:
            converter = UnitConverter.load(user=user, key=key, id=cp.batch_id)
        except ConverterLoadError:
            converter = UnitConverter(
                id=cp.batch_id,
                base_system=cp.base_system,
                base_unit=cp.base_unit,
                user=user,
                key=key
            )
        except UnitConverterInitError:
            return Response("Error initializing converter", status=status.HTTP_400_BAD_REQUEST)
        if cp.data:
            errors = converter.add_data(data=cp.data)
            if errors:
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = converter.convert()
            serializer = ConverterResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': converter.id, 'status': converter.status},
                            content_type="application/json")


class CustomUnitViewSet(ModelViewSet):
    """
    Custom Units API
    """
    queryset = CustomUnit.objects.all()
    serializer_class = CustomUnitSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = CustomUnitFilter
    pagination_class = PageNumberPagination
    permission_classes = [CustomUnitObjectPermission, permissions.IsAuthenticated]
    display_page_controls = True
    lookup_url_param = 'system_name'

    user = openapi.Parameter('user',
                             openapi.IN_QUERY,
                             description="Filter on user rates",
                             type=openapi.TYPE_BOOLEAN)
    key = openapi.Parameter('key',
                            openapi.IN_QUERY,
                            description="Filter on user defined category",
                            type=openapi.TYPE_STRING)
    unit_system = openapi.Parameter('unit_system',
                                    openapi.IN_QUERY,
                                    description="Filter on unit system",
                                    type=openapi.TYPE_STRING)
    code = openapi.Parameter('code',
                             openapi.IN_QUERY,
                             description="Filter on unit code",
                             type=openapi.TYPE_STRING)
    name = openapi.Parameter('name',
                             openapi.IN_QUERY,
                             description="Filter on unit name",
                             type=openapi.TYPE_STRING)
    relation = openapi.Parameter('relation',
                                 openapi.IN_QUERY,
                                 description="Filter on relation to base units",
                                 type=openapi.TYPE_STRING)
    symbol = openapi.Parameter('symbol',
                               openapi.IN_QUERY,
                               description="Filter on unit symbol",
                               type=openapi.TYPE_STRING)
    alias = openapi.Parameter('alias',
                              openapi.IN_QUERY,
                              description="Filter on unit alias",
                              type=openapi.TYPE_STRING)
    ordering = openapi.Parameter('ordering',
                                 openapi.IN_QUERY,
                                 description="Sort on code, name, relation, symbol, alias. "
                                             "Prefix with - for descending sort",
                                 type=openapi.TYPE_STRING)

    def get_queryset(self):
        """
        Filter units based on authenticated user
        """
        qs = super(CustomUnitViewSet, self).get_queryset()
        system_name = self.kwargs.get(self.lookup_url_param, 'SI')
        if self.request.user and self.request.user.is_authenticated:
            qs = qs.filter(
                models.Q(user=self.request.user) | models.Q(user__isnull=True)
            )
            if self.request.GET.get('key'):
                qs = qs.filter(key=self.request.GET.get('key'))
        else:
            qs = qs.filter(models.Q(user__isnull=True))
        qs = qs.filter(unit_system__iexact=system_name.lower())
        return qs

    @swagger_auto_schema(manual_parameters=[
        user, key, unit_system, code, name, relation, symbol, alias, ordering],
        responses={200: CustomUnitSerializer})
    def list(self, request, *args, **kwargs):
        """
        List rates
        """
        return super().list(request, *args, **kwargs)

    def create(self, request: HttpRequest, system_name: str, *args, **kwargs):
        """
        Create CustomUnit
        """
        cu_form = CustomUnitForm(request.data)
        if cu_form.is_valid():
            cu = cu_form.save(commit=False)
            try:
                us = UnitSystem(system_name=system_name, user=request.user, key=cu.key)
            except UnitSystemNotFound:
                return Response("Invalid unit system", status=status.HTTP_400_BAD_REQUEST)
            if request.user and request.user.is_authenticated:
                if CustomUnit.objects.filter(code=cu.code, user=request.user, key=cu.key).exists():
                    return Response("Custom unit already exists", status=status.HTTP_409_CONFLICT)
                cu.user = request.user
                cu.unit_system = system_name
                try:
                    cu.save()
                except (UnitValueError, ValueError) as e:
                    return Response(str(e), status=status.HTTP_400_BAD_REQUEST)
                serializer = CustomUnitSerializer(cu)
                return Response(serializer.data, status=status.HTTP_201_CREATED)
            else:
                return HttpResponseForbidden()
        else:
            return Response(cu_form.errors, status=status.HTTP_400_BAD_REQUEST,
                            content_type="application/json")


class ValidateViewSet(APIView):
    """
    POST API View to validate formulas
    """

    @swagger_auto_schema(request_body=ExpressionSerializer)
    @action(['POST'], detail=False, url_path='', url_name="validate")
    def post(self, request, unit_system, *args, **kwargs):
        """
        Validate a formula with parameters
        :param request: HTTP request
        :param unit_system: Unit system to use for validation
        """
        if request.user and request.user.is_authenticated:
            us = UnitSystem(unit_system, user=request.user, key=request.data.get('key', None))
        else:
            us = UnitSystem(unit_system)
        exp = ExpressionSerializer(data=request.data)
        if exp.is_valid(unit_system=us, dimensions_only=True):
            return Response("Valid expression")
        else:
            return Response(json.dumps(exp.errors),
                            status=status.HTTP_406_NOT_ACCEPTABLE,
                            content_type="application/json")


class CalculationView(APIView):
    """
    Calculate a batch of formulas
    """
    renderer_classes = [UnitsRenderer]

    @swagger_auto_schema(request_body=CalculationPayloadSerializer,
                         responses={200: CalculationResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency and date to a reference currency
        :param request: HTTP request
        """
        cps = CalculationPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST,
                            content_type="application/json")
        cp = cps.create(cps.validated_data)
        user = None
        if request.user and request.user.is_authenticated:
            user = request.user
        key = request.POST.get('key', None)
        try:
            calculator = ExpressionCalculator.load(user=user, key=key, id=cp.batch_id)
        except ConverterLoadError:
            calculator = ExpressionCalculator(
                id=cp.batch_id,
                unit_system=cp.unit_system,
                user=user,
                key=key
            )
        except ExpressionCalculatorInitError:
            return Response("Error initializing calculator", status=status.HTTP_400_BAD_REQUEST)
        if cp.data:
            errors = calculator.add_data(data=cp.data)
            if errors:
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = calculator.convert()
            serializer = CalculationResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': calculator.id, 'status': calculator.status},
                            content_type="application/json")



"""
Units models
"""

import logging
import re
from datetime import date

import pint.systems
import uncertainties
from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import ugettext as _
from sympy import sympify, SympifyError

from geocurrency.converters.models import BaseConverter, ConverterResult, \
    ConverterResultDetail, ConverterResultError, ConverterLoadError
from . import UNIT_EXTENDED_DEFINITION, DIMENSIONS, UNIT_SYSTEM_BASE_AND_DERIVED_UNITS, \
    ADDITIONAL_BASE_UNITS, PREFIX_SYMBOL
from .exceptions import *
from .settings import ADDITIONAL_UNITS, PREFIXED_UNITS_DISPLAY


class Quantity:
    """
    Quantity class
    """
    system = None
    unit = None
    value = 0
    date_obj = None

    def __init__(self, system: str, unit: str, value: float, date_obj: date = None):
        """
        Initialize quantity on unit system
        """
        self.system = system
        self.unit = unit
        self.value = value
        self.date_obj = date_obj

    def __repr__(self):
        """
        Look beautiful
        """
        return f'{self.value} {self.unit} ({self.system})'


class Unit:
    """
    Unit mock for hinting
    """
    pass


class UnitSystem:
    """
    Pint UnitRegistry wrapper
    """
    ureg = None
    system_name = None
    system = None
    _additional_units = set()

    def __init__(self, system_name: str = 'SI', fmt_locale: str = 'en', user: User = None,
                 key: str = None):
        """
        Initialize UnitSystem from name and user / key information for loading custom units
        """
        found = False
        for available_system in UnitSystem.available_systems():
            if system_name.lower() == available_system.lower():
                system_name = available_system
                found = True
        if not found:
            raise UnitSystemNotFound("Invalid unit system")
        self.system_name = system_name
        try:
            additional_units_settings = settings.GEOCURRENCY_ADDITIONAL_UNITS
        except AttributeError:
            additional_units_settings = ADDITIONAL_UNITS
        try:
            self.ureg = pint.UnitRegistry(system=system_name, fmt_locale=fmt_locale)
            self.system = getattr(self.ureg.sys, system_name)
            self._load_additional_units(units=ADDITIONAL_BASE_UNITS)
            self._load_additional_units(units=additional_units_settings)
            if user:
                self._load_custom_units(user=user, key=key)
            self._rebuild_cache()
        except (FileNotFoundError, AttributeError):
            raise UnitSystemNotFound("Invalid unit system")

    def _rebuild_cache(self):
        """
        Rebuild registry cache
        It should be in the define method of the registry
        """
        self.ureg._build_cache()

    def _load_additional_units(self, units: dict, redefine: bool = False) -> bool:
        """
        Load additional base units in registry
        """
        available_units = self.available_unit_names()
        if self.system_name not in units:
            logging.warning(f"error loading additional units for {self.system_name}")
            return False
        added_units = []
        for key, items in units[self.system_name].items():
            if key not in available_units:
                self.ureg.define(f"{key} = {items['relation']} = {items['symbol']}")
                added_units.append(key)
            elif redefine:
                self.ureg.redefine(f"{key} = {items['relation']} = {items['symbol']}")
        self._additional_units = self._additional_units | set(added_units)
        return True

    def _load_custom_units(self, user: User, key: str = None, redefine: bool = False) -> bool:
        """
        Load custom units in registry
        """
        if user and user.is_authenticated:
            if user.is_superuser:
                qs = CustomUnit.objects.all()
            else:
                qs = CustomUnit.objects.filter(user=user)
            if key:
                qs = qs.filter(key=key)
        else:
            qs = CustomUnit.objects.filter(pk=-1)
        qs = qs.filter(unit_system=self.system_name)
        available_units = self.available_unit_names()
        added_units = []
        for cu in qs:
            props = [cu.code, cu.relation]
            if cu.symbol:
                props.append(cu.symbol)
            if cu.alias:
                props.append(cu.alias)
            definition = " = ".join(props)
            if cu.code not in available_units:
                self.ureg.define(definition)
                added_units.append(cu.code)
            elif redefine:
                self.ureg.redefine(definition)
            else:
                logging.error(f"{cu.code} already defined in registry")
        self._additional_units = self._additional_units | set(added_units)
        return True

    def _test_additional_units(self, units: dict) -> bool:
        """
        Load and check dimensionality of ADDITIONAL_BASE_UNITS values
        """
        if self.system_name not in units:
            return False
        for key in units[self.system_name].keys():
            try:
                self.unit(key).dimensionality and True
            except pint.errors.UndefinedUnitError:
                return False
        return True

    def add_definition(self, code, relation, symbol, alias):
        """
        Add a new unit definition to a UnitSystem, and rebuild cache
        :param code: code of the unit
        :param relation: relation to other units (e.g.: 3 kg/m)
        :param symbol: short unit representation
        :param alias: other name for unit
        """
        self.ureg.define(f"{code} = {relation} = {symbol} = {alias}")
        self._rebuild_cache()

    @classmethod
    def available_systems(cls) -> [str]:
        """
        List of available Unit Systems
        :return: Array of string
        """
        ureg = pint.UnitRegistry(system='SI')
        return dir(ureg.sys)

    @classmethod
    def is_valid(cls, system: str) -> bool:
        """
        Check validity of the UnitSystem
        :param system: name of the unit system
        """
        us = cls()
        return system in us.available_systems()

    def current_system(self) -> pint.UnitRegistry:
        """
        Return current pint.UnitRegistry
        """
        return self.ureg

    def unit(self, unit_name):
        """
        Create a Object in the UnitSystem
        :param unit_name: name of the unit in the unit system
        """
        return Unit(unit_system=self, code=unit_name)

    def available_unit_names(self) -> [str]:
        """
        List of available units for a given Unit system
        :return: Array of names of Unit systems
        """
        try:
            prefixed_units_display = settings.GEOCURRENCY_PREFIXED_UNITS_DISPLAY
        except AttributeError:
            prefixed_units_display = PREFIXED_UNITS_DISPLAY
        prefixed_units = []
        for key, prefixes in prefixed_units_display.items():
            for prefix in prefixes:
                prefixed_units.append(prefix + key)
        return sorted(prefixed_units + dir(getattr(self.ureg.sys, self.system_name))
                      + list(self._additional_units))

    def unit_dimensionality(self, unit: str) -> str:
        """
        User friendly representation of the dimension
        :param unit: name of the unit to display
        :return: Human readable dimension
        """
        return Unit.dimensionality_string(unit_system=self.system, unit_str=unit)

    def available_dimensions(self, ordering: str = 'name') -> {}:
        """
        Return available dimensions for the UnitSystem
        :param ordering: sort result by attribute
        """
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['code', 'name', 'dimension']:
            ordering = 'name'
        return sorted([Dimension(unit_system=self, code=dim) for dim in DIMENSIONS.keys()],
                      key=lambda x: getattr(x, ordering, ''), reverse=descending)

    @property
    def _ureg_dimensions(self):
        """
        return dimensions with units
        """
        dimensions = []
        for dim in self.ureg._dimensions:
            try:
                if not self.ureg.get_compatible_units(dim):
                    continue
                dimensions.append(dim)
            except KeyError:
                continue
        return dimensions

    def _get_dimension_dimensionality(self, dimension: str) -> {}:
        """
        Return the dimensionality of a dimension based on the first compatible unit
        """
        try:
            for dim in self.ureg.get_compatible_units(dimension):
                return self.ureg.get_base_units(dim)[1]
        except KeyError:
            return {}

    def _generate_dimension_delta_dictionnary(self) -> {}:
        """
        Generate the dict to put in DIMENSIONS
        """
        output = {}
        for dim in self._ureg_dimensions:
            if dim not in DIMENSIONS:
                output[dim] = {
                    'name': f'_({dim})',
                    'dimension': str(self._get_dimension_dimensionality(dim)),
                    'symbol': ''
                }
        return output

    def units_per_dimension(self, dimensions: [str]) -> {}:
        """
        Return units grouped by dimension
        :param dimensions: restrict list of dimensions
        """
        output = {}
        registry_dimensions = dimensions or DIMENSIONS.keys()
        for dim in registry_dimensions:
            dimension = Dimension(unit_system=self, code=dim)
            try:
                units = self.ureg.get_compatible_units(dim)
                if units:
                    output[dim] = units
            except KeyError:
                continue
        return output

    def units_per_dimensionality(self) -> {}:
        """
        List of units per dimension
        :return: dict of dimensions, with lists of unit strings
        """
        units_array = self.available_unit_names()
        output = {}
        for unit_str in units_array:
            dimension = Unit.dimensionality_string(self, unit_str)
            try:
                output[dimension].append(unit_str)
            except KeyError:
                output[dimension] = [unit_str]
        return output

    @property
    def dimensionalities(self) -> [str]:
        """
        List of dimensions available in the Unit system
        :return: list of dimensions for Unit system
        """
        return set([Unit.dimensionality_string(self, unit_str) for unit_str in dir(self.system)])


class Dimension:
    """
    Dimenion of a Unit
    """
    unit_system = None
    code = None
    name = None
    dimension = None

    def __init__(self, unit_system: UnitSystem, code: str):
        """
        Initialize a Dimension in a UnitSystem
        """
        try:
            dimension = DIMENSIONS[code]
            self.unit_system = unit_system
            self.code = code
            self.name = dimension['name']
            self.dimension = dimension['dimension']
        except (ValueError, KeyError) as e:
            logging.warning(str(e))
            self.code = None
        if not self.code:
            raise DimensionNotFound

    def __repr__(self):
        """
        Look beautiful
        """
        return self.code

    def units(self, user=None, key=None) -> [Unit]:
        """
        List of units for this dimension
        :param user: optional user for custom units
        :param key: optional key for custom units
        """
        if self.code == '[compounded]':
            return self._compounded_units
        if self.code == '[custom]':
            return self._custom_units(user=user, key=key)
        unit_list = []
        try:
            prefixed_units_display = settings.GEOCURRENCY_PREFIXED_UNITS_DISPLAY
        except AttributeError:
            prefixed_units_display = PREFIXED_UNITS_DISPLAY
        try:
            unit_list.append(
                self.unit_system.unit(
                    UNIT_SYSTEM_BASE_AND_DERIVED_UNITS[self.unit_system.system_name][self.code]
                )
            )
        except (KeyError, UnitNotFound):
            logging.warning(f"unable to find base unit for"
                            f"unit system {self.unit_system.system_name}"
                            f" and dimension {self.code}")
        try:
            unit_list.extend(
                [
                    Unit(unit_system=self.unit_system, pint_unit=unit)
                    for unit in self.unit_system.ureg.get_compatible_units(self.code)
                ])
        except KeyError as e:
            logging.warning(f"Cannot find compatible units for this dimension {self.code}")
        unit_names = [str(u) for u in unit_list]
        for unit, prefixes in prefixed_units_display.items():
            if unit in unit_names:
                for prefix in prefixes:
                    unit_list.append(self.unit_system.unit(unit_name=prefix + unit))
        return set(sorted(unit_list, key=lambda x: x.name))

    @property
    def _compounded_units(self):
        """
        List units that do not belong to a dimension
        """
        available_units = self.unit_system.available_unit_names()
        dimensioned_units = []
        for dimension_code in [d for d in DIMENSIONS.keys() if
                               d != '[compounded]' and d != '[custom]']:
            dimension = Dimension(unit_system=self.unit_system, code=dimension_code)
            dimensioned_units.extend([u.code for u in dimension.units()])
        return [self.unit_system.unit(au) for au in set(available_units) - set(dimensioned_units)]

    def _custom_units(self, user: User, key: str = None) -> [Unit]:
        """
        Return list of custom units
        :param user: User owning the units
        :param key: optional unit key
        """
        if user and user.is_authenticated:
            if user.is_superuser:
                custom_units = CustomUnit.objects.all()
            else:
                custom_units = CustomUnit.objects.filter(user=user)
            if key:
                custom_units = custom_units.filter(key=key)
            return [self.unit_system.unit(cu.code) for cu in custom_units]
        else:
            return []

    @property
    def base_unit(self):
        """
        Base unit for this dimension in this Unit System
        """
        try:
            return UNIT_SYSTEM_BASE_AND_DERIVED_UNITS[self.unit_system.system_name][self.code]
        except KeyError:
            logging.warning(
                f'dimension {self.dimension} is not part of '
                f'unit system {self.unit_system.system_name}')
            return None


class Unit:
    """
    Pint Unit wrapper
    """
    unit_system = None
    code = None
    unit = None

    def __init__(self, unit_system: UnitSystem, code: str = '', pint_unit: pint.Unit = None):
        """
        Initialize a Unit in a UnitSystem
        :param unit_system: UnitSystem instance
        :param code: code of the pint.Unit
        """
        self.unit_system = unit_system
        if pint_unit and isinstance(pint_unit, pint.Unit):
            self.code = str(pint_unit)
            self.unit = pint_unit
        elif code:
            self.code = code
            try:
                self.unit = getattr(unit_system.system, code)
            except pint.errors.UndefinedUnitError:
                raise UnitNotFound("invalid unit for system")
        else:
            raise UnitNotFound("invalid unit for system")

    def __repr__(self):
        return self.code

    @classmethod
    def is_valid(cls, name: str) -> bool:
        """
        Check the validity of a unit in a UnitSystem
        """
        try:
            us_si = UnitSystem(system_name='SI')
        except UnitSystemNotFound:
            return False
        try:
            return us_si.unit(unit_name=name) and True
        except pint.errors.UndefinedUnitError:
            return False

    @property
    def name(self) -> str:
        """
        Return name of the unit from table of units
        """
        return self.unit_name(self.code)

    @property
    def symbol(self) -> str:
        """
        Return symbol for Unit
        """
        return self.unit_symbol(self.code)

    @property
    def dimensions(self) -> [Dimension]:
        """
        Return Dimensions of Unit
        """
        dimensions = [Dimension(unit_system=self.unit_system, code=code) for code in
                      DIMENSIONS.keys()
                      if DIMENSIONS[code]['dimension'] == str(self.dimensionality)]
        return dimensions or '[compounded]'

    @staticmethod
    def base_unit(unit_str: str) -> (str, str):
        """
        Get base unit in case the unit is a prefixed unit
        :param unit_str: name of unit to check
        :return: base unit name, prefix
        """
        prefix = ''
        base_str = unit_str
        try:
            prefixed_units_display = settings.GEOCURRENCY_PREFIXED_UNITS_DISPLAY
        except AttributeError:
            prefixed_units_display = PREFIXED_UNITS_DISPLAY
        for base, prefixes in prefixed_units_display.items():
            for _prefix in prefixes:
                if unit_str == _prefix + base:
                    prefix = _prefix
                    base_str = base
        return base_str, prefix

    @staticmethod
    def unit_name(unit_str: str) -> str:
        """
        Get translated name from unit string
        :param unit_str: Name of unit
        """
        base_str, prefix = Unit.base_unit(unit_str=unit_str)
        try:
            ext_unit = UNIT_EXTENDED_DEFINITION.get(base_str)
            return prefix + str(ext_unit['name'])
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {base_str}')
            return unit_str

    @staticmethod
    def unit_symbol(unit_str: str) -> str:
        """
        Static function to get symbol from unit string
        :param unit_str: Name of unit
        """
        base_str, prefix = Unit.base_unit(unit_str=unit_str)
        try:
            prefix_symbol = PREFIX_SYMBOL[prefix]
            ext_unit = UNIT_EXTENDED_DEFINITION.get(base_str)
            return prefix_symbol + ext_unit['symbol']
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {base_str}')
            return ''

    @staticmethod
    def dimensionality_string(unit_system: UnitSystem, unit_str: str) -> str:
        """
        Converts pint dimensionality string to human readable string
        :param unit_system: UnitSystem
        :param unit_str: Unit name
        :return: str
        """
        ds = str(getattr(unit_system.ureg, unit_str).dimensionality).replace('[', '').replace(']',
                                                                                              '')
        ds = ds.replace(' ** ', '^')
        ds = ds.split()
        return ' '.join([_(d) for d in ds])

    @property
    def dimensionality(self):
        """
        Return dimensionality of a unit in Pint universe
        """
        try:
            return self.unit_system.ureg.get_base_units(self.code)[1]
        except KeyError:
            return ''

    @staticmethod
    def translated_name(unit_system: UnitSystem, unit_str: str) -> str:
        """
        Translated name of the unit
        """
        try:
            return '{}'.format(unit_system.ureg[unit_str])
        except KeyError:
            return unit_str

    @property
    def readable_dimension(self):
        """
        Wrapper around Unit.dimensionality_string
        """
        return Unit.dimensionality_string(unit_system=self.unit_system, unit_str=self.code)


class UnitConverter(BaseConverter):
    """
    Conversion between units
    """
    base_system = None
    base_unit = None
    user = None
    key = None

    def __init__(self, base_system: str, base_unit: str, user: User = None, key: key = None,
                 id: str = None):
        """
        Initialize the converter. It converts a payload into a destination unit
        """
        try:
            super().__init__(id=id)
            self.base_system = base_system
            self.base_unit = base_unit
            self.user = user
            self.key = key
            self.system = UnitSystem(system_name=base_system, user=user, key=key)
            self.unit = Unit(unit_system=self.system, code=base_unit)
        except (UnitSystemNotFound, UnitNotFound):
            raise UnitConverterInitError

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super().add_data(data)
        return errors

    def check_data(self, data):
        """
        Validates that the data contains
        system = str
        unit = str
        value = float
        date_obj ('YYYY-MM-DD')
        """
        from .serializers import QuantitySerializer
        errors = []
        for line in data:
            serializer = QuantitySerializer(data=line)
            if serializer.is_valid():
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    @classmethod
    def load(cls, id: str, user: User = None, key: str = None) -> BaseConverter:
        """
        Load converter from ID
        """
        try:
            uc = super().load(id)
            uc.system = UnitSystem(system_name=uc.base_system, user=user, key=key)
            uc.unit = Unit(unit_system=uc.system, code=uc.base_unit)
            return uc
        except (UnitSystemNotFound, UnitNotFound, KeyError) as e:
            raise ConverterLoadError

    def save(self):
        """
        Save the converter to cache
        """
        system = self.system
        unit = self.unit
        self.system = None
        self.unit = None
        super().save()
        self.system = system
        self.unit = unit

    def convert(self) -> ConverterResult:
        """
        Converts data to base unit in base system
        """

        result = ConverterResult(id=self.id, target=self.base_unit)
        q_ = self.system.ureg.Quantity
        for quantity in self.data:
            try:
                pint_quantity = q_(quantity.value, quantity.unit)
                out = pint_quantity.to(self.base_unit)
                result.increment_sum(out.magnitude)
                detail = ConverterResultDetail(
                    unit=quantity.unit,
                    original_value=quantity.value,
                    date=quantity.date_obj,
                    conversion_rate=0,
                    converted_value=out.magnitude
                )
                result.detail.append(detail)
            except pint.UndefinedUnitError:
                error = ConverterResultError(
                    unit=quantity.unit,
                    original_value=quantity.value,
                    date=quantity.date_obj,
                    error=_('Undefined unit in the registry')
                )
                result.errors.append(error)
            except pint.DimensionalityError:
                error = ConverterResultError(
                    unit=quantity.unit,
                    original_value=quantity.value,
                    date=quantity.date_obj,
                    error=_('Dimensionality error, incompatible units')
                )
                result.errors.append(error)
        self.end_batch(result.end_batch())
        return result


class UnitConversionPayload:
    """
    Unit conversion payload
    """
    data = None
    base_system = ''
    base_unit = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(self, base_system: UnitSystem, base_unit: Unit, data=None, key: str = None,
                 batch_id: str = None, eob: bool = False):
        """
        Initialize conversion payload
        """
        self.data = data
        self.base_system = base_system
        self.base_unit = base_unit
        self.key = key
        self.batch_id = batch_id
        self.eob = eob


class CustomUnit(models.Model):
    """
    Additional unit for a user
    """
    AVAILABLE_SYSTEMS = (
        ('Planck', 'Planck'),
        ('SI', 'SI'),
        ('US', 'US'),
        ('atomic', 'atomic'),
        ('cgs', 'CGS'),
        ('imperial', 'imperial'),
        ('mks', 'mks'),
    )
    user = models.ForeignKey(User, related_name='units', on_delete=models.PROTECT)
    key = models.CharField("Categorization field (e.g.: customer ID)",
                           max_length=255, default=None, db_index=True, null=True, blank=True)
    unit_system = models.CharField("Unit system to register the unit in", max_length=20,
                                   choices=AVAILABLE_SYSTEMS)
    code = models.SlugField("technical name of the unit (e.g.: myUnit)")
    name = models.CharField("Human readable name (e.g.: My Unit)", max_length=255)
    relation = models.CharField("Relation to an existing unit (e.g.: 12 kg*m/s)", max_length=255)
    symbol = models.CharField("Symbol to use in a formula (e.g.: myu)", max_length=20, blank=True,
                              null=True)
    alias = models.CharField("Other code for this unit (e.g.: mybu)", max_length=20, null=True,
                             blank=True)

    class Meta:
        """
        Meta
        """
        unique_together = ('user', 'key', 'code')
        ordering = ['name', 'code']

    def save(self, *args, **kwargs):
        """
        Save custom unit to database
        """
        us = UnitSystem(system_name=self.unit_system)
        self.code = self.code.replace('-', '_')
        self.symbol = self.symbol.replace('-', '_')
        self.alias = self.alias.replace('-', '_')
        if self.code in us.available_unit_names():
            raise UnitDuplicateError
        try:
            us.add_definition(code=self.code, relation=self.relation, symbol=self.symbol,
                              alias=self.alias)
        except ValueError as e:
            raise UnitValueError(str(e)) from e
        try:
            us.unit(self.code).unit.dimensionality
        except pint.errors.UndefinedUnitError:
            raise UnitDimensionError
        return super(CustomUnit, self).save(*args, **kwargs)


class Operand:
    """
    Operand in a formula
    """
    name = None
    value = None
    unit = None
    uncertainty = None

    def __init__(self,
                 name: str = None,
                 value: float = 0,
                 unit: str = None,
                 uncertainty: str = None):
        """
        Initialize Operand
        :param name: name of the operand that appears in the formula
        :param value: float magnitude
        :param unit: units
        """
        self.name = name
        self.value = value
        self.unit = unit
        self.uncertainty = uncertainty
        if self.validate():
            self.uvalue = uncertainties.ufloat(value, self.parse_uncertainty())
        else:
            self.uvalue = uncertainties.ufloat(0, 0)

    def validate(self):
        """
        Validate operand
        """
        if not self.name or self.value is None or self.unit is None:
            return False
        try:
            float(self.value)
        except ValueError:
            return False
        try:
            self.parse_uncertainty()
        except (ValueError, AttributeError):
            return False
        return True

    def get_unit(self, unit_system: UnitSystem) -> str:
        """
        Transform unit if unit is a dimension
        """
        # Find dimensions
        exp = r'(?P<dim>\[\w+\])'
        replace_dict = {}
        for dimension_name in re.findall(exp, self.unit):
            try:
                dim = Dimension(unit_system=unit_system, code=dimension_name)
                replace_dict[dimension_name] = str(dim.units().pop())
            except DimensionNotFound:
                pass
        # Replace in formula
        for key, item in replace_dict.items():
            self.unit = self.unit.replace(key, item)
        return self.unit

    def get_magnitude(self):
        return self.uvalue.n

    def get_uncertainty(self):
        return self.uvalue.s

    def parse_uncertainty(self):
        if not self.uncertainty:
            return 0
        if self.uncertainty.endswith('%'):
            return self.value * float(self.uncertainty[:-1]) * 0.01
        else:
            return float(self.uncertainty)


class ComputationError(Exception):
    pass


class Expression:
    """
    Expression with operands
    """
    expression = None
    operands = None
    out_units = None

    def __init__(self, expression: str, operands: [Operand], out_units: str = None):
        """
        Initialize Expression
        :param expression: string expression with placeholders
        :param operands: List of Operand corresponding to placeholders
        """
        self.expression = expression
        self.operands = operands
        self.out_units = out_units

    def validate(self, unit_system: UnitSystem) -> (bool, str):
        """
        Validate syntax and homogeneity
        :param unit_system: UnitSystem
        """
        q_ = unit_system.ureg.Quantity
        if not self.expression:
            return False, "missing expression"
        try:
            sympify(self.expression.format(**{v.name: v.value for v in self.operands}))
        except SympifyError as e:
            return False, "Improper expression"
        for var in self.operands:
            if not var.validate():
                return False, "invalid operand"
        kwargs = {v.name: q_(v.value, v.unit) for v in self.operands}
        try:
            result = unit_system.ureg.parse_expression(self.expression, **kwargs)
        except KeyError:
            return False, "Missing operand"
        except pint.errors.DimensionalityError:
            return False, "Incoherent dimensions"
        if self.out_units:
            try:
                result.to(self.out_units)
            except pint.errors.DimensionalityError:
                return False, "Incoherent output dimensions"
        return True, ''

    def evaluate(self, unit_system: UnitSystem) -> pint.Quantity:
        """
        Validate formula
        """
        q_ = unit_system.ureg.Quantity
        is_valid, error = self.validate(unit_system=unit_system)
        if is_valid:
            kwargs = {v.name: q_(v.uvalue, v.unit) for v in self.operands}
            result = unit_system.ureg.parse_expression(self.expression, **kwargs)
            if self.out_units:
                return result.to(self.out_units)
            else:
                return result
        else:
            raise ComputationError(f"Invalid formula: {error}")


class CalculationPayload:
    """
    Calculation payload: caculate expressions
    """
    data = None
    unit_system = ''
    user = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(self, unit_system: UnitSystem, key: str = None, data: [] = None,
                 batch_id: str = None, eob: bool = False):
        """
        Initialize payload
        """
        self.data = data
        self.unit_system = unit_system
        self.key = key
        self.batch_id = batch_id
        self.eob = eob


class CalculationResultDetail:
    """
    Details of an evaluation
    """
    expression = None
    operands = None
    magnitude = None
    uncertainty = None
    unit = None

    def __init__(self, expression: str, operands: [],
                 magnitude: uncertainties.ufloat, unit: str):
        """
        Initialize detail
        :param expression: Expression in the form of a string e.g.: 3*{a} + 2 * {b}
        :param operands: List of Operands
        :param magnitude: result of the calculation
        :param units: dimension of the result
        :param uncertainty: uncertainty of the calculation
        """
        self.expression = expression
        self.operands = operands
        self.magnitude = magnitude.n
        self.uncertainty = magnitude.s
        self.unit = unit


class CalculationResultError:
    """
    Error details of a wrong calculation
    """
    expression = None
    operands = None
    calc_date = None
    error = None

    def __init__(self, expression: str, operands: [], calc_date: date, error: str):
        """
        Initialize error detail
        :param expression: Expression in the form of a string
        :param operands: List of Operands
        :param date: date of the evaluation
        :param error: Error description
        """
        self.expression = expression
        self.operands = operands
        self.calc_date = calc_date
        self.error = error


class CalculationResult:
    """
    Result of a batch of evaluations
    """
    id = None
    detail = []
    status = None
    errors = []

    def __init__(self, id: str = None, detail: [CalculationResultDetail] = None,
                 status: str = BaseConverter.INITIATED_STATUS,
                 errors: [CalculationResultError] = None):
        """
        Initialize result
        :param id: ID of the batch
        :param detail: List of CalculationResultDetail
        :param status: status of the batch
        :param errors: List of CalculationResultErrors
        """
        self.id = id
        self.detail = detail or []
        self.status = status
        self.errors = errors or []

    def end_batch(self):
        """
        Puts a finall status on the batch
        """
        if self.errors:
            self.status = BaseConverter.WITH_ERRORS
        else:
            self.status = BaseConverter.FINISHED
        return self.status


class ExpressionCalculator(BaseConverter):
    """
    Conversion between units
    """
    unit_system = None

    def __init__(self, unit_system: str, user: User = None, key: str = '', id: str = None):
        """
        Initiate ExpressionCalculator
        :param unit_system: unit system name
        :param user: User
        :param key: key of user
        :param id: ID of the batch
        """
        try:
            super().__init__(id=id)
            self.unit_system = unit_system
            self.user = user
            self.key = key
            self.system = UnitSystem(system_name=unit_system, user=self.user, key=self.key)
        except UnitSystemNotFound as e:
            raise ExpressionCalculatorInitError from e

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super().add_data(data)
        return errors

    def check_data(self, data):
        """
        Validates that the data contains
        - system (str)
        - unit (str)
        - value (float)
        - date (YYYY-MM-DD)
        """
        from .serializers import ExpressionSerializer
        errors = []
        for line in data:
            serializer = ExpressionSerializer(data=line)
            if serializer.is_valid(unit_system=self.system):
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    @classmethod
    def load(cls, id: str, user: User = None, key: str = None) -> BaseConverter:
        """
        Load converter from batch
        :param id: ID of the batch
        :param user: Users object
        :param key: User defined category
        """
        try:
            uc = super().load(id)
            uc.system = UnitSystem(system_name=uc.unit_system, user=user, key=key)
            return uc
        except (UnitSystemNotFound, KeyError) as e:
            raise ConverterLoadError

    def save(self):
        """
        Save converter to caching system
        """
        system = self.system
        self.system = None
        super(ExpressionCalculator, self).save()
        self.system = system

    def convert(self) -> CalculationResult:
        """
        Converts data to base unit in base system
        """
        result = CalculationResult(id=self.id)
        for expression in self.data:
            valid, exp_error = expression.validate(unit_system=self.system)
            if not valid:
                error = CalculationResultError(
                    expression=expression.expression,
                    operands=expression.operands,
                    calc_date=date.today(),
                    error=exp_error
                )
                result.errors.append(error)
                continue
            try:
                out = expression.evaluate(unit_system=self.system)
            except ComputationError as e:
                error = CalculationResultError(
                    expression=expression.expression,
                    operands=expression.operands,
                    calc_date=date.today(),
                    error=str(e)
                )
                result.errors.append(error)
                continue
            detail = CalculationResultDetail(
                expression=expression.expression,
                operands=expression.operands,
                magnitude=out.magnitude,
                unit=out.units
            )
            result.detail.append(detail)
        self.end_batch(result.end_batch())
        return result



# Generated by Django 3.2 on 2021-04-18 19:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rates', '0003_auto_20210318_0713'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rate',
            name='base_currency',
            field=models.CharField(db_index=True, default='EUR', max_length=3, verbose_name='Currency to convert to'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='currency',
            field=models.CharField(db_index=True, max_length=3, verbose_name='Currency to convert from'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='key',
            field=models.CharField(db_index=True, default=None, max_length=255, null=True, verbose_name='User defined categorization key'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='value',
            field=models.FloatField(default=0, verbose_name='Rate conversion factor'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='value_date',
            field=models.DateField(verbose_name='Date of value'),
        ),
    ]



"""
Country views
"""

import logging
import os

import requests
from django.conf import settings
from django.http import HttpResponseBadRequest, HttpResponseNotFound
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from sendfile import sendfile

from .models import Country, CountryNotFoundError
from .settings import *


class FlagView(View):
    """
    Try to get flag image from MEDIA_ROOT cache or download it from FLAG_SOURCE
    """

    @method_decorator(cache_page(60 * 60 * 2))
    @method_decorator(vary_on_cookie)
    def get(self, request, pk, *args, **kwargs):
        """
        Handle GET request
        """
        try:
            country = Country(alpha_2=pk)
            flag_path = os.path.join(settings.MEDIA_ROOT, country.alpha_2 + '.svg')
            if not os.path.exists(flag_path):
                response = requests.get(FLAG_SOURCE.format(alpha_2=country.alpha_2))
                try:
                    flag_content = response.text
                    flag_file = open(flag_path, 'w')
                    flag_file.write(flag_content)
                    flag_file.close()
                except IOError:
                    logging.error("unable to write file", flag_path)
                    return HttpResponseBadRequest("Error fetching file")
            return sendfile(request, flag_path)
        except CountryNotFoundError as e:
            logging.error("Error fetching country")
            logging.error(e)
            return HttpResponseNotFound("Invalid country")



"""
Converter views
"""

from django.http import HttpResponseNotFound
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.views import APIView

from .models import BaseConverter, Batch
from .serializers import BatchSerializer


class WatchView(APIView):
    """
    View to track orgression of a conversion batch
    """

    @swagger_auto_schema(responses={200: BatchSerializer})
    @action(['GET'], detail=True, url_path='', url_name="watch")
    def get(self, request, converter_id, *args, **kwargs):
        """
        GET handler
        :param request: HTTPRequest
        :param converter_id: ID of the converter
        :param args: list of arguments
        :param kwargs: dict of arguments
        """
        try:
            converter = BaseConverter.load(converter_id)
        except KeyError:
            return HttpResponseNotFound('Converter not found')
        batch = Batch(converter_id, converter.status)
        serializer = BatchSerializer(batch)
        return Response(serializer.data, content_type="application/json")



"""
Serializers for Rats module
"""
from datetime import datetime, date

from geocurrency.core.serializers import UserSerializer
from rest_framework import serializers

from .models import Rate, Amount, BulkRate, RateConversionPayload


class BulkSerializer(serializers.Serializer):
    """
    Serializer for Bulk serializer
    """
    base_currency = serializers.CharField(label="Currency to convert to", max_length=3,
                                          required=True)
    currency = serializers.CharField(label="Currency to convert from", max_length=3, required=True)
    value = serializers.FloatField(label="rate value", required=True)
    key = serializers.CharField(label="User defined categorization ID", required=True)
    from_date = serializers.DateField(label="Date to create the rate from", required=True)
    to_date = serializers.DateField(label="Date to create the rate to", required=False)

    def create(self, validated_data):
        """
        Create a BulkRate object
        """
        return BulkRate(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a BulkRate object
        """
        instance.base_currency = validated_data.get('base_currency', instance.base_currency)
        instance.currency = validated_data.get('currency', instance.currency)
        instance.value = validated_data.get('value', instance.value)
        instance.key = validated_data.get('key', instance.key)
        instance.from_date = validated_data.get('key', instance.from_date)
        instance.to_date = validated_data.get('key', instance.to_date)
        return instance

    @staticmethod
    def validate_currency(value):
        """
        validate currency
        :param value: Currency
        """
        from geocurrency.currencies.models import Currency
        if not Currency.is_valid(value):
            raise serializers.ValidationError('Invalid currency')
        return value

    @staticmethod
    def validate_base_currency(value):
        """
        Validate base currency
        :param value: Currency
        """
        from geocurrency.currencies.models import Currency
        if not Currency.is_valid(value):
            raise serializers.ValidationError('Invalid currency')
        return value

    @staticmethod
    def validate_from_date(value):
        """
        Validate from_date
        :param value: date
        """
        if isinstance(value, date):
            return value
        try:
            datetime.strptime(value, 'YYYY-MM-DD')
        except ValueError:
            raise serializers.ValidationError('Invalid date format, use YYYY-MM-DD')
        return value

    @staticmethod
    def validate_to_date(value):
        """
        Validate to_date
        :param value: date
        """
        if isinstance(value, date):
            return value
        try:
            datetime.strptime(value, 'YYYY-MM-DD')
        except ValueError:
            raise serializers.ValidationError('Invalid date format, use YYYY-MM-DD')
        return value


class RateSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField(label="ID of the rate")
    user = UserSerializer(label="Owner of the rate", read_only=True)

    class Meta:
        model = Rate
        fields = [
            'id',
            'user',
            'key',
            'currency',
            'base_currency',
            'value_date',
            'value',
        ]


class RateStatItemSerializer(serializers.Serializer):
    """
    Rate statistics item for conversion
    """
    currency = serializers.CharField(label="Currency to convert from", read_only=True)
    base_currency = serializers.CharField(label="Currency to convert to", read_only=True)
    period = serializers.CharField(label="Period over which to aggregate", read_only=True)
    avg = serializers.FloatField(label="Average value", read_only=True)
    max = serializers.FloatField(label="Max value", read_only=True)
    min = serializers.FloatField(label="Min value", read_only=True)
    std_dev = serializers.FloatField(label="Standard deviation", read_only=True)


class RateStatSerializer(serializers.Serializer):
    """
    Rate statistics used in conversion
    """
    key = serializers.CharField(label="User defined categorization key", read_only=True)
    period = serializers.CharField(label="Period over which to aggregate", read_only=True)
    from_date = serializers.DateField(label="start aggregation date", read_only=True)
    to_date = serializers.DateField(label="end aggregation date", read_only=True)
    results = RateStatItemSerializer(label="List of results", many=True, read_only=True)


class RateAmountSerializer(serializers.Serializer):
    """
    Rate amount used in conversion
    """
    currency = serializers.CharField(label="Currency of the amount")
    amount = serializers.FloatField(label="Value to convert")
    date_obj = serializers.DateField(label="Date of conversion")

    @staticmethod
    def validate_currency(value):
        """
        validate currency
        :param value: Currency
        """
        from geocurrency.currencies.models import Currency
        if not Currency.is_valid(value):
            raise serializers.ValidationError('Invalid currency')
        return value

    @staticmethod
    def validate_amount(value):
        """
        validate amount
        :param value: Amount
        """
        try:
            float(value)
        except ValueError:
            raise serializers.ValidationError('Invalid number')
        return value

    @staticmethod
    def validate_date_obj(value):
        """
        Validate date
        :param value: date
        """
        if isinstance(value, date):
            return value
        try:
            datetime.strptime(value, 'YYYY-MM-DD')
        except ValueError:
            raise serializers.ValidationError('Invalid date format, use YYYY-MM-DD')
        return value

    def create(self, validated_data: dict) -> Amount:
        """
        Create Amount object
        :param validated_data: cleaned data
        """
        return Amount(**validated_data)

    def update(self, instance: Amount, validated_data: dict):
        """
        Update Amount object
        :param instance: Amount
        :param validated_data: cleaned date
        """
        instance.currency = validated_data.get('currency', instance.email)
        instance.amount = validated_data.get('amount', instance.content)
        instance.date_obj = validated_data.get('date_obj', instance.created)
        return instance


class RateConversionPayloadSerializer(serializers.Serializer):
    """
    Serialize a conversion payload
    """
    data = RateAmountSerializer(label="Amounts to convert", many=True, required=False)
    target = serializers.CharField(label="Target currency", required=True)
    batch_id = serializers.CharField(label="User defined batch ID", required=False)
    key = serializers.CharField(label="User defined categorization key", required=False)
    eob = serializers.BooleanField(label="End of batch? Triggers the conversion", default=False)

    def is_valid(self, raise_exception=False):
        """
        Check validity of the payload
        """
        if not self.initial_data.get('data') and (not self.initial_data.get('batch_id')
                                                  or (self.initial_data.get('batch_id')
                                                      and not self.initial_data.get('eob'))):
            raise serializers.ValidationError(
                'data has to be provided if batch_id '
                'is not provided or batch_id is provided and eob is False'
            )
        return super(RateConversionPayloadSerializer, self).is_valid()

    @staticmethod
    def validate_target(value: str):
        """
        Validate target currency
        :param value: string
        """
        from geocurrency.currencies.models import Currency
        if not Currency.is_valid(value):
            raise serializers.ValidationError('Invalid currency')
        return value

    def create(self, validated_data: dict) -> RateConversionPayload:
        """
        Create a RateConversionPayload from serialized data
        :param validated_data: cleaned data
        """
        return RateConversionPayload(**validated_data)

    def update(self, instance, validated_data):
        """
        Update RateConversionPayload from serialized data
        :param instance: RateConversionPayload object
        :param validated_data: cleaned data
        """
        self.data = validated_data.get('data', instance.data)
        self.target = validated_data.get('target', instance.target)
        self.batch_id = validated_data.get('batch_id', instance.batch_id)
        self.key = validated_data.get('key', instance.key)
        self.eob = validated_data.get('eob', instance.eob)
        return instance



"""
List of serializers for Converters and Calculators
"""

from rest_framework import serializers

from .models import Batch, ConverterResultError, ConverterResultDetail, ConverterResult







class ConverterResultDetailSerializer(serializers.Serializer):
    """
    Serialize a ConverterResultDetail class
    """
    unit = serializers.CharField(label="Unit of conversion")
    original_value = serializers.FloatField(label="Original value to convert")
    date = serializers.DateField(label="Date of conversion")
    conversion_rate = serializers.FloatField(label="Conversion rate")
    converted_value = serializers.FloatField(label="Resulting value")

    def create(self, validated_data):
        """
        Create a ConverterResultDetail object
        :param validated_data: cleaned data
        """
        return ConverterResultDetail(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a ConverterResultDetail object
        :param instance: ConverterResultDetail object
        :param validated_data: cleaned data
        """
        instance.unit = validated_data.get('unit', instance.unit)
        instance.original_value = validated_data.get('original_value', instance.original_value)
        instance.date = validated_data.get('date', instance.date)
        return instance


class ConverterResultErrorSerializer(serializers.Serializer):
    """
    Serializer for ConverterResultError
    """
    unit = serializers.CharField(label="Unit of conversion")
    original_value = serializers.FloatField(label="Original value")
    date = serializers.DateField(label="Date of conversion")
    error = serializers.CharField(label="Error during conversion")

    def create(self, validated_data):
        """
        Create a ConverterResultError object
        :param validated_data: cleaned data
        """
        return ConverterResultError(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a ConverterResultError object
        :param instance: ConverterResultError object
        :param validated_data: cleaned data
        """
        instance.unit = validated_data.get('unit', instance.unit)
        instance.original_value = validated_data.get('original_value', instance.original_value)
        instance.date = validated_data.get('date', instance.date)
        instance.error = validated_data.get('error', instance.error)
        return instance


class ConverterResultSerializer(serializers.Serializer):
    """
    Serializer for a ConverterResult
    """
    id = serializers.UUIDField(label="ID of the batch")
    target = serializers.CharField(label="Target conversion")
    detail = ConverterResultDetailSerializer(label="Details of conversion", many=True)
    sum = serializers.FloatField(label="Sum of conversions")
    status = serializers.CharField(label="Status of the conversion")
    errors = ConverterResultErrorSerializer(label="Errors during conversions", many=True)

    def create(self, validated_data):
        """
        Create a ConverterResult object
        :param validated_data: cleaned data
        """
        return ConverterResult(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a ConverterResult object
        :param instance: ConverterResult object
        :param validated_data: cleaned data
        """
        instance.id = validated_data.get('id', instance.id)
        instance.target = validated_data.get('target', instance.target)
        instance.detail = validated_data.get('detail', instance.detail)
        instance.sum = validated_data.get('sum', instance.sum)
        instance.status = validated_data.get('status', instance.status)
        instance.errors = validated_data.get('errors', instance.errors)
        return instance


class BatchSerializer(serializers.Serializer):
    """
    Serializer for Batch
    """
    id = serializers.CharField(label="Batch unique identifier")
    status = serializers.CharField(label="Status of the batch")

    def create(self, validated_data):
        """
        Create a Batch object
        :param validated_data: cleaned data
        """
        return Batch(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Batch
        :param instance: Batch object
        :param validated_data: cleaned data
        """
        instance.status = validated_data.get('status', instance.status)
        return instance



"""
Units module URLs
"""

from django.conf.urls import url, include
from django.urls import path
from rest_framework import routers

from .viewsets import UnitSystemViewset, UnitViewset, ConvertView, CustomUnitViewSet, \
    ValidateViewSet, CalculationView

app_name = 'units'

router = routers.DefaultRouter()
router.register(r'', UnitSystemViewset, basename='unit_systems')
router.register(r'(?P<system_name>\w+)/units', UnitViewset, basename='units')
router.register(r'(?P<system_name>\w+)/custom', CustomUnitViewSet, basename='custom')

urlpatterns = [
    path('convert/', ConvertView.as_view()),
    path('<str:unit_system>/formulas/validate/', ValidateViewSet.as_view()),
    path('<str:unit_system>/formulas/calculate/', CalculationView.as_view()),
    url(r'^', include(router.urls)),
]



"""
Python_forex based wrapper
uses ratesapi.io
"""
import logging
from datetime import date, timedelta
from typing import Iterator

from django.conf import settings
from forex_python import converter

from . import RateService, RatesNotAvailableError


class ForexService(RateService):
    """
    Forex Rate service
    """

    def available_currencies(self) -> Iterator:
        """
        List available currencies for this service
        """
        rates = self.fetch_rates(base_currency='USD')
        return [r['currency'] for r in rates]

    def _fetch_all_rates(self, provider: converter.CurrencyRates, base_currency: str,
                         date_obj: date) -> []:
        """
        Fetch all rates for availbale currencies for this service
        :param provider: Currency rates
        :param base_currency: base currency
        :param date_obj: date of value
        :return: List of conversion rates
        """
        try:
            _rates = provider.get_rates(base_cur=base_currency, date_obj=date_obj)
            return [self.serializer(
                base_currency=base_currency,
                currency=key,
                date=date_obj,
                value=value
            ) for key, value in _rates.items()]
        except converter.RatesNotAvailableError as e:
            logging.error(e)
            raise RatesNotAvailableError

    def _fetch_single_rate(self,
                           provider: converter.CurrencyRates,
                           base_currency: str,
                           currency: str, date_obj: date) -> []:
        """
        Fetch one conversion rate between a currency and a base currency at a given date
        :param provider: provider of the values
        :param base_currency: base currency
        :param currency: currency to convert from
        :param date_obj: date of value
        :return: List of rates
        """
        try:
            value = provider.get_rate(dest_cur=currency, base_cur=base_currency, date_obj=date_obj)
            return [self.serializer(
                base_currency=base_currency,
                currency=currency,
                date=date_obj,
                value=value
            )]

        except converter.RatesNotAvailableError as e:
            logging.error(e)
            raise RatesNotAvailableError

    def fetch_rates(self,
                    base_currency: str = settings.BASE_CURRENCY,
                    currency: str = None,
                    date_obj: date = date.today(), to_obj: date = None) -> []:
        """
        Get conversion rates between currency and base currency for a range of dates
        :param base_currency: currency to convert to
        :param currency: currency to convert from
        :param date_obj: beginning of range
        :param to_obj: end of range
        :return: List of conversion rates
        """
        c = converter.CurrencyRates()
        rates = []
        _rates = []
        if currency:
            for i in range(((to_obj or date_obj) - date_obj).days + 1):
                _rates = self._fetch_single_rate(c, base_currency=base_currency, currency=currency,
                                                 date_obj=date_obj + timedelta(i))
                rates.extend(_rates)
        else:
            for i in range(((to_obj or date_obj) - date_obj).days + 1):
                _rates = self._fetch_all_rates(c, base_currency=base_currency,
                                               date_obj=date_obj + timedelta(i))
                rates.extend(_rates)
        return rates



"""
Settings for Currencies module
"""

RATES = {
    'EUR': {
        'source': {
            'name': 'BCE',
            'url': 'https://www.ecb.europa.eu/stats/policy_and_exchange_rates'
                   '/euro_reference_exchange_rates/html/index.en.html'
        },
        'history': 'https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.xml',
        'current': 'https://www.ecb.europa.eu/stats/eurofxref/eurofxref-daily.xml'
    },
    # 'CAD': {
    #     'source': {
    #       'name': 'Bank of Canada',
    #       'url': 'https://www.bankofcanada.ca/rates/exchange/daily-exchange-rates/'
    #     },
    #     'history': 'https://www.bankofcanada.ca/valet/observations
    #     /group/FX_RATES_DAILY/csv?start_date=2017-01-03',
    #     'current': 'https://www.bankofcanada.ca/valet/observations
    #     /group/FX_RATES_DAILY/csv?start_date=2017-01-03'
    # },
}

BASE_CURRENCY = 'EUR'



"""
Serializers for Units module
"""

import json
import logging
from datetime import date, datetime

import pint
from drf_yasg.utils import swagger_serializer_method
from rest_framework import serializers
from sympy import sympify, SympifyError

from geocurrency.core.serializers import UserSerializer
from .models import Quantity, UnitConversionPayload, Dimension, \
    CustomUnit, Expression, CalculationPayload, UnitSystem, Operand, Unit, \
    CalculationResult, CalculationResultError, CalculationResultDetail


class QuantitySerializer(serializers.Serializer):
    """
    Serialize a Quantity
    """
    system = serializers.CharField(label="Unit system of the Quantity")
    unit = serializers.CharField(label="Units of the quantity")
    value = serializers.FloatField(label="Magnitude of the quantity")
    date_obj = serializers.DateField(label="date of conversion")

    @staticmethod
    def validate_system(system):
        """
        Validate unit system by name
        :param system: system name
        """
        from .models import UnitSystem
        if not UnitSystem.is_valid(system):
            raise serializers.ValidationError('Invalid system')
        return system

    @staticmethod
    def validate_unit(unit):
        """
        Validating unit requires knowledge of the system that might embark user information for
        custom units
        Let‘s say it is valid for now
        """
        return unit

    @staticmethod
    def validate_date_obj(value):
        """
        Validate date
        :param value:
        """
        if isinstance(value, date):
            return value
        try:
            datetime.strptime(value, 'YYYY-MM-DD')
        except ValueError:
            raise serializers.ValidationError('Invalid date format, use YYYY-MM-DD')
        return value

    def create(self, validated_data):
        """
        Create a Quantity object
        :param validated_data: cleaned data
        """
        return Quantity(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Quantity object
        :param instance: Quantity object
        :param validated_data: cleaned data
        """
        instance.system = validated_data.get('system', instance.system)
        instance.unit = validated_data.get('unit', instance.unit)
        instance.value = validated_data.get('value', instance.value)
        instance.date_obj = validated_data.get('date_obj', instance.date_obj)
        return instance


class UnitSerializer(serializers.Serializer):
    """
    Unit class serializer
    """
    code = serializers.CharField(label="Technical name of the unit")
    name = serializers.CharField(label="Human readable name of the unit")
    dimensions = serializers.SerializerMethodField(label="Dimensions the unit belongs to")

    @swagger_serializer_method(serializer_or_field=serializers.ListField)
    def get_dimensions(self, obj):
        """
        Get dimension of unit
        :param obj: Unit instance
        """
        return str(obj.dimensions)


class DimensionSerializer(serializers.Serializer):
    """
    Dimension class serializer
    """
    code = serializers.CharField(label="technical name of the unit (enclosed in brackets)")
    name = serializers.CharField(label="Human readable name of the unit")
    dimension = serializers.CharField(label="Mathematical expression of the dimension")
    base_unit = serializers.SerializerMethodField(label="Name of the base unit")

    def create(self, validated_data):
        """
        Create a Dimension object from cleaned data
        :param validated_data: cleaned data
        """
        return Dimension(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Dimension object
        :param instance: Dimension object
        :param validated_data: cleaned data
        """
        instance.unit_system = validated_data.get('system', instance.unit_system)
        instance.code = validated_data.get('code', instance.code)
        instance.value = validated_data.get('name', instance.name)
        instance.dimension = validated_data.get('date_obj', instance.dimension)
        return instance

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_base_unit(self, obj: Dimension) -> Unit:
        """
        Get base unit for dimension object
        :param obj: Dimension object
        """
        return obj.base_unit


class DimensionWithUnitsSerializer(DimensionSerializer):
    """
    Serialize a Dimension including associated units
    """
    units = UnitSerializer(label="Units of this dimension", many=True)

    @swagger_serializer_method(serializer_or_field=UnitSerializer)
    def get_units(self, obj: Dimension) -> [Unit]:
        """
        Get units for this dimension
        """
        try:
            return obj.units()
        except KeyError as e:
            logging.error(str(e))
            return []


class UnitSystemSerializer(serializers.Serializer):
    """
    Serialize a UnitSystem
    """
    system_name = serializers.CharField()


class UnitConversionPayloadSerializer(serializers.Serializer):
    """
    Serialize a UnitConversionPayload
    """
    data = QuantitySerializer(label="List of unit conversions", many=True, required=False)
    base_system = serializers.CharField(label="Unit system used for conversion", required=True)
    base_unit = serializers.CharField(label="Unit to express the result in", required=True)
    batch_id = serializers.CharField(label="User defined batch ID", required=False)
    key = serializers.CharField(label="User defined categorization key", required=False)
    eob = serializers.BooleanField(label="End of batch ? triggers the conversion", default=False)
    _errors = {}

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of a converson payload
        """
        super().is_valid(raise_exception=raise_exception)
        if not self.initial_data.get('data') and (not self.initial_data.get('batch_id')
                                                  or (self.initial_data.get('batch_id')
                                                      and not self.initial_data.get('eob'))):
            self._errors = {
                'data': 'data has to be provided if batch_id is not provided'
                        'or batch_id is provided and eob is False'
            }

        if self._errors and raise_exception:
            raise serializers.ValidationError(self.errors)
        return not bool(self._errors)

    @staticmethod
    def validate_base_system(value: str):
        """
        Validate unit system
        :param value: unit system name
        """
        from geocurrency.units.models import UnitSystem
        if not UnitSystem.is_valid(value):
            raise serializers.ValidationError('Invalid unit system')
        return value

    @staticmethod
    def validate_base_unit(value: str) -> str:
        """
        Validate base unit
        :param value: name of unit
        """
        from geocurrency.units.models import Unit
        if not Unit.is_valid(value):
            raise serializers.ValidationError('Invalid unit')
        return value

    def create(self, validated_data: {}) -> UnitConversionPayload:
        """
        Create a UnitConversionPayload
        :param validated_data: cleaned data
        """
        return UnitConversionPayload(**validated_data)

    def update(self, instance, validated_data: {}) -> UnitConversionPayload:
        """
        Update a UnitConversionPayload
        :param instance: UnitConversionPayload
        :param validated_data: cleaned data
        """
        self.data = validated_data.get('data', instance.data)
        self.base_system = validated_data.get('base_system', instance.base_system)
        self.base_unit = validated_data.get('base_system', instance.base_unit)
        self.batch_id = validated_data.get('batch_id', instance.batch_id)
        self.key = validated_data.get('key', instance.key)
        self.eob = validated_data.get('eob', instance.eob)
        return instance


class CustomUnitSerializer(serializers.ModelSerializer):
    """
    Serialize a custom unit
    """
    id = serializers.CharField(label="ID of the CustomUnit", read_only=True)
    user = UserSerializer(label="Owner of the unit", read_only=True)
    unit_system = serializers.CharField(label="Unit system of the unit", read_only=True)

    class Meta:
        """
        Meta
        """
        model = CustomUnit
        fields = ['id', 'user', 'key', 'unit_system', 'code', 'name', 'relation', 'symbol', 'alias']


class OperandSerializer(serializers.Serializer):
    """
    Serialize an Operand
    """
    name = serializers.CharField(label="Name of the operand in the expression",
                                 help_text="{a} in an expression should have an operand named 'a'")
    value = serializers.FloatField(label="Value of the operand as a float")
    unit = serializers.CharField(label="Units of the operand as a string")
    uncertainty = serializers.CharField(label="Uncertainty value, either absolute, or percentage.",
                                        required=False,
                                         help_text="Uncertainty for value, "
                                                   "if float, it is absolute (10), "
                                                   "else, it must contain % (12%)")

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of the operand
        """
        if not self.initial_data.get('name'):
            raise serializers.ValidationError("operand Name not set")
        if 'value' not in self.initial_data:
            raise serializers.ValidationError("operand Value not set")
        try:
            float(self.initial_data['value'])
        except ValueError as e:
            raise serializers.ValidationError("invalid operand value") from e
        if 'unit' not in self.initial_data:
            raise serializers.ValidationError("operand Unit not set")
        return super().is_valid(raise_exception=raise_exception)

    def create(self, validated_data):
        """
        Create an Operand object
        :param validated_data: cleaned date
        """
        return Operand(**validated_data)


class ExpressionListSerializer(serializers.ListSerializer):
    """
    Serialize a list of Expressions
    """

    def is_valid(self, unit_system: UnitSystem, dimensions_only=False, raise_exception=False):
        """
        Check validity of expression with a specific UnitSystem
        :param unit_system: UnitSystem instance
        :param dimensions_only: Only check dimensionality, not actual units
        :param raise_exception: raise an exception instead of a list of errors
        """
        super().is_valid(raise_exception=raise_exception)
        for initial_exp in self.initial_data:
            ec = ExpressionSerializer(data=initial_exp)
            ec.is_valid(unit_system=unit_system, dimensions_only=dimensions_only)
            if ec.errors:
                self._errors.append(ec.errors)
        return not bool(self._errors)


class ExpressionSerializer(serializers.Serializer):
    """
    Serialize an expression
    """
    unit_system = None
    key = None
    expression = serializers.CharField(label="Mathematical expression to evaluate", required=True,
                                       help_text="Operands must be placed between curly braces."
                                                 "A recommended expression format "
                                                 "would be '{a}+{b}*{c}'")
    operands = OperandSerializer(label="List of operands", many=True, required=True)
    out_units = serializers.CharField(label="Define output units", required=False,
                                      help_text="Set output units, optional (e.g.: km/hour)")

    class Meta:
        """
        Meta
        """
        list_serializer_class = ExpressionListSerializer

    def is_valid(self, unit_system: UnitSystem,
                 raise_exception=False,
                 dimensions_only=False) -> bool:
        """
        Check if expression is valid
        :param unit_system: UnitSystem instance
        :param dimensions_only: Only test dimensions
        :param raise_exception: raise an exception instead of a list of errors
        """
        super().is_valid(raise_exception=raise_exception)
        operands = self.operands_validation(
            operands=self.initial_data.get('operands')
        )
        expression = self.expression_validation(
            expression=self.initial_data.get('expression'),
            operands=operands
        )
        out_units = self.out_units_validation(
            unit_system=unit_system,
            out_units=self.initial_data.get('out_units')
        )
        if expression:
            if dimensions_only:
                self.dimensions_validation(
                    unit_system=unit_system,
                    expression=expression,
                    operands=operands,
                    out_units=out_units
                )
            else:
                self.units_validation(
                    unit_system=unit_system,
                    expression=expression,
                    operands=operands,
                    out_units=out_units
                )
        if self._errors and raise_exception:
            raise serializers.ValidationError(self._errors)
        return not bool(self._errors)

    def units_validation(self,
                         unit_system: UnitSystem,
                         expression: str,
                         operands: [Operand],
                         out_units: str = None) -> bool:
        """
        Validate units based on the units of the operands
        :param unit_system: Reference unit system with custom units
        :param expression: Mathematical expression as string
        :param operands: list of Operand objects
        :param out_units: optional conversion to specific units
        """
        # Validate units
        q_ = unit_system.ureg.Quantity
        try:
            units_kwargs = {v['name']: q_(v['value'], v['unit']) for v in operands}
            result = unit_system.ureg.parse_expression(expression, **units_kwargs)
        except KeyError as e:
            self._errors['operands'] = "Missing operands"
            return False
        except pint.errors.DimensionalityError:
            self._errors['expression'] = "Incoherent dimension"
            return False
        if out_units:
            try:
                result.to(out_units)
            except pint.errors.DimensionalityError:
                self._errors['out_units'] = "Incoherent output dimension"
                return False
        return True

    def dimensions_validation(self,
                              unit_system: UnitSystem,
                              expression: str,
                              operands: [Operand],
                              out_units: str = None) -> bool:
        """
        Validate units dimensions based on get_unit of the operands
        :param unit_system: Reference unit system with custom units
        :param expression: Mathematical expression as string
        :param operands: list of Operand objects
        :param out_units: optional conversion to specific units
        """
        q_ = unit_system.ureg.Quantity
        operand_objs = [Operand(name=v['name'], value=v['value'], unit=v['unit']) for v in operands]
        units_kwargs = {v.name: f"{v.value} {v.get_unit(unit_system)}" for v in operand_objs}
        try:
            result = q_(expression.format(**units_kwargs))
        except KeyError:
            self._errors['operands'] = "Missing operands"
            return False
        except pint.errors.DimensionalityError:
            self._errors['expression'] = "Incoherent dimension"
            return False
        if out_units:
            try:
                result.to(out_units)
            except pint.errors.DimensionalityError:
                self._errors['out_units'] = "Incoherent output dimension"
                return False
        return True

    def expression_validation(self, expression: str, operands: [Operand]):
        """
        Check syntactic validity of expression
        :param expression: Mathematical expression as string
        :param operands: list of Operand objects
        """
        if not expression:
            self._errors['expression'] = "Empty expression"
        # Validate syntax
        value_kwargs = {v['name']: v['value'] for v in operands}
        try:
            sympify(expression.format(**value_kwargs))
        except (SympifyError, KeyError):
            self._errors['expression'] = f"Invalid operation"
        return expression

    def out_units_validation(self, unit_system: UnitSystem, out_units: str):
        """
        Check validity of output units
        :param unit_system: Reference unit system with custom units
        :param out_units: optional conversion to specific units
        """
        if not out_units:
            return None
        q_ = unit_system.ureg.Quantity
        try:
            q_(1, out_units)
        except pint.errors.DimensionalityError:
            self._errors['out_units'] = "Incoherent output dimension"
            return None
        return out_units

    def operands_validation(self, operands):
        """
        Validate Operand
        :param operands: Operand
        """
        errors = []
        try:
            for var in operands:
                vs = OperandSerializer(data=var)
                vs.is_valid()
                if vs.errors:
                    errors.append(vs.errors)
            if errors:
                self._errors['operands'] = errors
        except json.JSONDecodeError as e:
            self._errors['operands'] = f"Invalid operands json format: {e}"
        return operands

    def create(self, validated_data):
        """
        Create an Expression
        :param validated_data: cleaned data
        """
        operands = []
        for var in validated_data.get('operands'):
            vs = OperandSerializer(data=var)
            if vs.is_valid():
                operands.append(vs.create(vs.validated_data))
        return Expression(
            expression=validated_data.get('expression'),
            operands=operands,
            out_units=validated_data.get('out_units')
        )

    def update(self, instance, validated_data):
        """
        Update Expression
        :param instance: Expression object
        :param validated_data: cleaned data
        """
        instance.expression = validated_data.get('expression')
        instance.operands = validated_data.get('operands')
        instance.out_units = validated_data.get('out_units')
        return instance


class CalculationResultDetailSerializer(serializers.Serializer):
    """
    Serializer for the CalculationResultDetail class
    """
    expression = serializers.CharField(label="Expression to evaluate")
    operands = OperandSerializer(many=True)
    magnitude = serializers.FloatField(label="Magnitude of result")
    uncertainty = serializers.FloatField(label="Uncertainty of result")
    unit = serializers.CharField(label="Units of result")

    def create(self, validated_data):
        """
        Create a CalculationResultDetail object
        :param validated_data: cleaned data
        """
        return CalculationResultDetail(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a CalculationResultDetail object
        :param instance: CalculationResultDetail object
        :param validated_data: cleaned data
        """
        instance.expression = validated_data.get('expression', instance.expression)
        instance.operands = validated_data.get('operands', instance.operands)
        instance.magnitude = validated_data.get('magnitude', instance.magnitude)
        instance.uncertainty = validated_data.get('uncertainty', instance.operands)
        return instance


class CalculationResultErrorSerializer(serializers.Serializer):
    """
    Serializer for the CalculationResultError class
    """
    expression = serializers.CharField(label="Expression to evaluate")
    operands = OperandSerializer(many=True)
    calc_date = serializers.DateField(label="Date of calculation")
    error = serializers.CharField(label="Error during calculation")

    def create(self, validated_data):
        """
        Create a CalculationResultError object
        :param validated_data: cleaned data
        """
        return CalculationResultError(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a CalculationResultError object
        :param instance: CalculationResultError object
        :param validated_data: cleaned data
        """
        instance.expression = validated_data.get('expression', instance.expression)
        instance.operands = validated_data.get('operands', instance.operands)
        instance.calc_date = validated_data.get('calc_date', instance.calc_date)
        instance.error = validated_data.get('error', instance.error)
        return instance


class CalculationResultSerializer(serializers.Serializer):
    id = serializers.UUIDField(label="ID of the batch")
    detail = CalculationResultDetailSerializer(label="Details of the calculation", many=True)
    status = serializers.CharField(label="Status of the calculation")
    errors = CalculationResultErrorSerializer(label="Errors during calculation", many=True)

    def create(self, validated_data):
        """
        Create a CalculationResult object
        :param validated_data: cleaned data
        """
        return CalculationResult(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a CalculationResult object
        :param instance: CalculationResult object
        :param validated_data: cleaned data
        """
        instance.id = validated_data.get('id', instance.id)
        instance.detail = validated_data.get('detail', instance.detail)
        instance.status = validated_data.get('status', instance.status)
        instance.errors = validated_data.get('errors', instance.errors)
        return instance


class CalculationPayloadSerializer(serializers.Serializer):
    """
    Serializer for CalculationPayload
    """
    unit_system = serializers.CharField(label="Unit system to evaluate in", required=True)
    data = ExpressionSerializer(label="Payload of expressions", many=True, required=False)
    batch_id = serializers.CharField(label="User defined ID of the batch of evaluations",
                                     required=False)
    key = serializers.CharField(label="User defined categorization key", required=False)
    eob = serializers.BooleanField(label="End of batch ? triggers the evaluation", default=False)

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of the payload
        """
        if not self.initial_data.get('data') and (not self.initial_data.get('batch_id')
                                                  or (self.initial_data.get('batch_id')
                                                      and not self.initial_data.get('eob'))):
            raise serializers.ValidationError(
                'data has to be provided if batch_id is not provided '
                'or batch_id is provided and eob is False'
            )
        return super().is_valid(raise_exception=raise_exception)

    @staticmethod
    def validate_unit_system(value: str):
        """
        Validate unit system
        :param value: unit system name
        """
        from geocurrency.units.models import UnitSystem
        if not UnitSystem.is_valid(value):
            raise serializers.ValidationError('Invalid unit system')
        return value

    def create(self, validated_data: {}):
        """
        Create CalculationPayload from cleaned data
        :param validated_data: cleaned data
        """
        return CalculationPayload(**validated_data)

    def update(self, instance, validated_data: {}):
        """
        Update a CalculationPayload from cleaned data
        :param instance: CalculationPayload object
        :param validated_data: cleaned data
        """
        self.data = validated_data.get('data', instance.data)
        self.unit_system = validated_data.get('unit_system', instance.unit_system)
        self.batch_id = validated_data.get('batch_id', instance.batch_id)
        self.key = validated_data.get('key', instance.key)
        self.eob = validated_data.get('eob', instance.eob)
        return instance



"""
REST filters for Rate APIs
"""

from django.db import models
from django.db.models import OuterRef, Subquery, QuerySet
from django_filters import rest_framework as filters

from .models import Rate


class RateFilter(filters.FilterSet):
    """
    Rate object filter
    """
    user = filters.BooleanFilter(label="filter rate associated to connected user",
                                 method='user_filter')
    key = filters.CharFilter(label="filter rates with key", method='key_filter')
    key_or_null = filters.CharFilter(label="filter rates with key or without key",
                                     method='key_or_null_filter')
    key_isnull = filters.CharFilter(label="filter rates without key", method='key_isnull_filter')
    value_date = filters.DateFilter(label="filter rates at a specific date",
                                    field_name='value_date', lookup_expr='exact')
    from_obj = filters.DateFilter(label="filter rates after a specific date (included)",
                                  field_name='value_date', lookup_expr='gte')
    to_obj = filters.DateFilter(label="filter rates before a specific date (included)",
                                field_name='value_date', lookup_expr='lte')
    value = filters.NumberFilter(label="filter rates with a specific value", field_name='value',
                                 lookup_expr='exact')
    lower_bound = filters.NumberFilter(
        label="filter rates with a value higher than the given value",
        field_name='value', lookup_expr='gte')
    higher_bound = filters.NumberFilter(
        label="filter rates with a value lower than the given value",
        field_name='value', lookup_expr='lte')
    currency = filters.CharFilter(label="filter by target currency", field_name='currency',
                                  lookup_expr='iexact')
    base_currency = filters.CharFilter(label="filter by base currency", field_name='base_currency',
                                       lookup_expr='iexact')
    currency_latest_values = filters.CharFilter(label="Only output latest rates for currency",
                                                method='currency_latest_values_filter')
    base_currency_latest_values = filters.CharFilter(label="Only output latest rates for currency",
                                                     method='base_currency_latest_values_filter')
    ordering = filters.OrderingFilter(
        # tuple-mapping retains order
        fields=(
            ('key', 'key'),
            ('value', 'value'),
            ('value_date', 'value_date'),
            ('base_currency', 'base_currency'),
            ('currency', 'currency'),
        ),
    )

    class Meta:
        """
        Meta
        """
        model = Rate
        exclude = ['pk', ]

    def user_filter(self, queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Filter on user
        """
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
            })
        return queryset.filter(user__isnull=True)

    def key_filter(self, queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Filter on key, only filters if request.user is set and authenticated
        """
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
                'key': value
            })
        return queryset.filter(user__isnull=True)

    def key_or_null_filter(self, queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Filter on key if user is authenticated or on records without user
        """
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(
                (models.Q(user=self.request.user) & models.Q(key=value)) | models.Q(
                    key__isnull=True)
            )
        return queryset.filter(user__isnull=True)

    @staticmethod
    def key_isnull_filter(queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Filter on records without key
        """
        return queryset.filter(key__isnull=True)

    @staticmethod
    def currency_latest_values_filter(queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Returns a queryset of latest values fos a currency
        """
        queryset = queryset.filter(currency=value)
        latest = queryset.filter(currency=OuterRef('currency')).order_by('-value_date')
        return queryset.annotate(
            currency_latest=Subquery(latest.values('value_date')[:1])
        ).filter(value_date=models.F('currency_latest'))

    @staticmethod
    def base_currency_latest_values_filter(queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Returns a queryset of latest valeus for a base currency
        """
        queryset = queryset.filter(base_currency=value)
        latest = queryset.filter(base_currency=OuterRef('base_currency')).order_by('-value_date')
        return queryset.annotate(
            base_currency_latest=Subquery(latest.values('value_date')[:1])
        ).filter(value_date=models.F('base_currency_latest'))



"""
Settings for Units module
"""
from django.conf import settings

# This will display these units in the list of units
ADDITIONAL_UNITS = {}

# Show units with prefixed values in APIs
PREFIXED_UNITS_DISPLAY = {
    'meter': ['milli', 'centi', 'kilo'],
    'gram': ['milli', 'kilo'],
    'second': ['micro', 'milli'],
    'ampere': ['milli'],
}



"""
Rates module URLs
"""

from django.conf.urls import url, include
from django.urls import path
from rest_framework import routers

from .viewsets import RateViewSet, ConvertView

app_name = 'rates'

router = routers.DefaultRouter()
router.register(r'', RateViewSet, basename='rates')

urlpatterns = [
    path('convert/', ConvertView.as_view()),
    url(r'^', include(router.urls)),
]



"""
Models for Country module
"""

import logging
import os
import re
from datetime import datetime

import pytz
import requests
from countryinfo import CountryInfo
from django.conf import settings
from django.core.cache import caches, cache
from django.db import models
from pycountry import countries
from pytz import timezone

from .helpers import ColorProximity, hextorgb
from .settings import *


class CountryNotFoundError(Exception):
    """
    Exception when Country is not found
    """
    msg = 'Country not found'


class CountryManager(models.Manager):
    """
    Manager for country model
    """

    @staticmethod
    def get_by_color(color, proximity=1):
        """
        Take a hex color and finds near country based on flag and color proximity
        :param color: hex color (FFF, FFFFFF, FFFFFFFF, #FFF, #FFFFFF, #FFFFFFFF)
        :param proximity: succes rate, positive if below (100 is opposite, 0 is identical
        """
        cp = ColorProximity()
        rgb_color = hextorgb(color)
        ctrs = []
        for country in Country.objects.filter(colors__isnull=False):
            for fc in country.colors.split(','):
                flag_color = hextorgb(fc)
                if cp.proximity(rgb_color, flag_color) < proximity:
                    ctrs.append(country.pk)
                    break
        return Country.objects.filter(pk__in=set(ctrs))


class Country:
    """
    Country class
    Wrapper around pycountry object
    """
    # data extracted from pycountry as basic data if CountryInfo for this country does not exist
    alpha_2 = None
    alpha_3 = None
    name = None
    numeric = None

    def __init__(self, alpha_2):
        """
        Init a Country object with an alpha2 code
        :params country_name: ISO-3166 alpha_2 code
        """
        country = countries.get(alpha_2=alpha_2)
        if not country:
            raise CountryNotFoundError("Invalid country alpha2 code")
        self.alpha_2 = country.alpha_2
        self.alpha_3 = country.alpha_3
        self.name = country.name
        self.numeric = country.numeric

    @classmethod
    def search(cls, term:str) -> []:
        """
        Search for Contruy by name, alpha_2, alpha_3, or numeric value
        :param term: Search term
        """
        result = []
        for attr in ['alpha_2', 'alpha_3', 'name', 'numeric']:
            result.extend(
                [getattr(c, 'alpha_2') for c in countries
                 if term.lower() in getattr(c, attr).lower()]
            )
        return sorted([Country(r) for r in set(result)], key=lambda x: x.name)

    @classmethod
    def all_countries(cls, ordering: str = 'name'):
        """
        List all countries, instanciate CountryInfo for each country in pycountry.countries
        :param ordering: sort list
        """
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['name', 'alpha_2', 'alpha_3', 'numeric']:
            ordering = 'name'
        return list(sorted(map(lambda x: cls(x.alpha_2), countries),
                           key=lambda x: getattr(x, ordering),
                           reverse=descending))

    def base(self):
        """
        Returns a basic representation of a country with name and iso codes
        """
        return countries.get(alpha_2=self.alpha_2)._fields

    def currencies(self) -> []:
        """
        Return a list of currencies used in this country
        """
        ci = CountryInfo(self.alpha_2)
        return ci.currencies()

    def info(self):
        """
        Additional information on the country based on countryinfo module database
        """
        return CountryInfo(self.alpha_2).info()

    @property
    def unit_system(self) -> str:
        """
        Return UnitSystem for country
        """
        if self.alpha_2 == 'US':
            return 'US'
        if self.alpha_2 == 'LR':
            return 'US'
        if self.alpha_2 == 'MM':
            return 'imperial'
        return 'SI'

    @property
    def timezones(self) -> []:
        """
        Returns a list of timezones for a country
        """
        output = []
        fmt = '%z'
        base_time = datetime.utcnow()
        for tz_info in pytz.country_timezones[self.alpha_2]:
            tz = timezone(tz_info)
            offset = tz.localize(base_time).strftime(fmt)
            numeric_offset = float(offset[:-2] + '.' + offset[-2:])
            output.append({
                'name': tz_info,
                'offset': f'UTC {offset}',
                'numeric_offset': numeric_offset,
                'current_time': base_time.astimezone(tz).strftime('%Y-%m-%d %H:%M')
            })
        return sorted(output, key=lambda x: x['numeric_offset'])

    @property
    def flag_path(self):
        """
        Path to the flag temporary file
        :return: string, absolute path to the flag file
        """
        return os.path.join(settings.MEDIA_ROOT, self.alpha_2 + '.svg')

    def flag_exists(self):
        """
        Checks if flag file exists
        :return: bool, True if flag exists, False otherwise
        """
        return os.path.exists(self.flag_path)

    def download_flag(self):
        """
        Downloads flag for country in temporary path
        :return: Path to the file
        """
        if not self.flag_exists():
            response = requests.get(FLAG_SOURCE.format(alpha_2=self.alpha_2))
            try:
                flag_content = response.text
                flag_file = open(self.flag_path, 'w')
                flag_file.write(flag_content)
                flag_file.close()
                return self.flag_path
            except IOError:
                logging.error(f"unable to write file {self.flag_path}")
                return None

    def analyze_flag(self):
        """
        Analyze colors of the flag for the country and caches the result
        :returns: array, list of colors
        """
        flag_path = os.path.join(settings.MEDIA_ROOT, self.alpha_2 + '.svg')
        # Checks if flag has been downloaded, downloads it otherwise,
        # and return None if download failed
        if not self.flag_exists() and not self.download_flag():
            return None
        with open(flag_path, 'r') as flag:
            content = flag.read()
            result = re.findall(r'\#[0-9A-Fa-f]{1,2}[0-9A-Fa-f]{1,2}[0-9A-Fa-f]{1,2}', content)
            if result:
                cache.set('COLORS-' + self.alpha_2, result)
            return result

    def colors(self):
        """
        List colors present in the flag of the country
        :returns: array, list of colors
        """
        colors = cache.get('COLORS-' + self.alpha_2)
        if colors:
            return colors
        else:
            return self.analyze_flag()

    @property
    def info(self) -> str:
        """
        Return country region
        """
        try:
            ccache = caches['countries']
        except KeyError:
            ccache = cache
        if not ccache.get(self.alpha_2, {}):
            try:
                info = CountryInfo(self.alpha_2).info()
            except KeyError:
                info = {}
            ccache.set(self.alpha_2, info)
        return ccache.get(self.alpha_2, {})

    @property
    def region(self) -> str:
        """
        Return country region
        """
        return self.info.get('region', '')

    @property
    def subregion(self) -> str:
        """
        Return country subregion
        """
        return self.info.get('subregion', '')

    @property
    def tld(self) -> str:
        """
        Return country TLD
        """
        return self.info.get('tld', '')

    @property
    def capital(self) -> str:
        """
        Return country capital
        """
        return self.info.get('capital', '')

    @property
    def population(self) -> str:
        """
        Return country population
        """
        return self.info.get('population', '')



"""
Permissions for CustomUnit APIs
"""

from rest_framework import permissions


class CustomUnitObjectPermission(permissions.BasePermission):
    """
    Permissions for CustomUnit API
    """

    def has_object_permission(self, request, view, obj):
        """
        Limit creation and modification tu logged in users
        """
        if not request.user or not request.user.is_authenticated:
            return False
        if request.method in permissions.SAFE_METHODS:
            return True
        if request.method == 'POST':
            return True
        elif request.method.lower() in ['put', 'patch', 'delete'] and request.user == obj.user:
            return True
        return False



"""
Country URLs
"""

from django.conf.urls import url, include
from rest_framework import routers

from .views import FlagView
from .viewsets import CountryViewset

app_name = 'countries'

router = routers.DefaultRouter()
router.register(r'', CountryViewset, basename='countries')

urlpatterns = [
    url(r'^', include(router.urls)),
    url(r'^(?P<pk>[^/.]+)/flag/$', FlagView.as_view())
]



"""
Permissions for Rate APIs
"""
from rest_framework import permissions


class RateObjectPermission(permissions.BasePermission):
    """
    Permissions for /rates APIs
    """

    def has_object_permission(self, request, view, obj):
        """
        Limit modifications to authenticated users
        """
        if request.method in permissions.SAFE_METHODS:
            return True
        if request.user and request.user.is_authenticated:
            if request.method == 'POST':
                return True
            elif request.method.lower() in ['put', 'patch', 'delete'] and request.user == obj.user:
                return True
        return False



"""
Units module API filters
"""

from django.db import models
from django.db.models import QuerySet
from django_filters import rest_framework as filters

from .models import CustomUnit


class CustomUnitFilter(filters.FilterSet):
    """
    Filter on custom units
    """
    user = filters.BooleanFilter(label="filter rate associated to connected user",
                                 method='user_filter')
    key = filters.CharFilter(label="filter rates with key", method='key_filter')
    unit_system = filters.CharFilter(label="filter by unit system", field_name='unit_system',
                                     lookup_expr='iexact')
    code = filters.CharFilter(label="filter by code", field_name='code', lookup_expr='iexact')
    name = filters.CharFilter(label="filter by name", field_name='name', lookup_expr='icontains')
    relation = filters.CharFilter(label="filter by relation", field_name='relation',
                                  lookup_expr='icontains')
    symbol = filters.CharFilter(label="filter by symbol", field_name='symbol', lookup_expr='iexact')
    alias = models.CharField("Alias", max_length=20, null=True, blank=True)

    ordering = filters.OrderingFilter(
        # tuple-mapping retains order
        fields=(
            ('key', 'key'),
            ('unit_system', 'unit_system'),
            ('code', 'code'),
            ('name', 'name'),
            ('relation', 'relation'),
            ('symbol', 'symbol'),
            ('alias', 'alias'),
        ),
    )

    class Meta:
        """
        Meta
        """
        model = CustomUnit
        fields = [
            'user', 'key',
            'unit_system', 'code', 'name',
            'relation', 'symbol', 'alias'

        ]

    def user_filter(self, queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Filter on request user
        """
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
            })
        return queryset.filter(user__isnull=True)

    def key_filter(self, queryset: QuerySet, name: str, value: str) -> QuerySet:
        """
        Filter on key if request.user is set and authenticated
        """
        if self.request and self.request.user and self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
                'key': value
            })
        return queryset.filter(user__isnull=True)



"""
Command to fetch rates for all currencies
"""
from datetime import datetime, date

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """
    Fetch rates command
    """
    help = 'Fetch rates for a date'

    def add_arguments(self, parser):
        """
        Add date argument to the command
        """
        parser.add_argument("-f", '--from_date', type=str,
                            help="Date in YYYY-MM-DD format."
                                 "Start fetching from this date, defaults to today")
        parser.add_argument("-t", '--to_date', type=str,
                            help="Date in YYYY-MM-DD format."
                                 "End of fetch range. defaults to today")
        parser.add_argument("-s", '--service', type=str,
                            help="Rate service name. (forex, currencylayer, or other)"
                                 "Defaults to forex. See SERVICES in settings")

    def handle(self, *args, **options):
        """
        Handle call
        """
        from geocurrency.currencies.models import Currency
        from geocurrency.rates.models import Rate
        today = date.today().strftime('%Y-%m-%d')
        try:
            from_date = datetime.strptime(options.get('from_date') or today, '%Y-%m-%d')
            to_date = datetime.strptime(options.get('to_date') or today, '%Y-%m-%d')
        except ValueError:
            print("invalid dates")
            exit(-1)
        rate_service = options.get('service') or 'forex'

        for currency in Currency.all_currencies():
            self.stdout.write('fetching rates for currency {}.'.format(currency.code))
            Rate.objects.fetch_rates(
                base_currency=currency.code,
                date_obj=from_date,
                to_obj=to_date,
                rate_service=rate_service
            )



"""
Units exception classes
"""


class UnitSystemNotFound(Exception):
    """
    UnitSystem not found exception
    """
    message = "Unit system name is not valid"


class DimensionNotFound(Exception):
    """
    Dimension not found exception
    """
    message = "Dimension not found in the registry"


class UnitNotFound(Exception):
    """
    Unit not found exception
    """
    message = "Unit not found in the registry"


class UnitDuplicateError(Exception):
    """
    Error duplicating unit
    """
    message = "Unit with the same code already exists in the registry"


class UnitValueError(Exception):
    """
    Error creating unit
    """
    message = "Unit values are inconsistent"


class UnitDimensionError(Exception):
    """
    Invalid dimension for the unit system
    """
    message = "Unit dimensionality is not valid for the registry"


class UnitConverterInitError(Exception):
    """
    Converter cannot be initialized
    """
    message = 'Error during initialization of converter'


class UnitConverterConvertError(Exception):
    """
    Error converting units
    """
    message = 'Error converting values'


class ExpressionCalculatorInitError(Exception):
    """
    Error initializing  Calculator
    """
    message = "Error initializing Expression calculator"



"""
Django settings for GeoCurrencies project.

Generated by 'django-admin startproject' using Django 3.0.5.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os

import pycountry

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = ''
# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

ALLOWED_HOSTS = [
    # Put your service URL
]

CORS_ORIGIN_ALLOW_ALL = False
CORS_ALLOW_CREDENTIALS = True

CORS_ORIGIN_WHITELIST = [
    # Put your service URL or front end URL
]

CORS_ALLOW_METHODS = [
    'DELETE',
    'GET',
    'OPTIONS',
    'PATCH',
    'POST',
    'PUT',
]

CORS_ALLOW_HEADERS = [
    'accept',
    'accept-encoding',
    'authorization',
    'content-type',
    'dnt',
    'origin',
    'user-agent',
    'x-csrftoken',
    'x-requested-with',
]

# Application definition

INSTALLED_APPS = [
    'channels',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_extensions',
    'rest_framework',
    'rest_framework.authtoken',
    'drf_yasg',
    'django_filters',
    'corsheaders',
    'geocurrency',
    'geocurrency.core',
    'geocurrency.countries',
    'geocurrency.currencies',
    'geocurrency.rates',
    'geocurrency.units',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

REST_FRAMEWORK = {
    'DEFAULT_FILTER_BACKENDS': ['django_filters.rest_framework.DjangoFilterBackend'],
    'DEFAULT_PAGINATION_CLASS': 'geocurrency.core.pagination.PageNumberPagination',
    'PAGE_SIZE': 100,
    'DEFAULT_THROTTLE_CLASSES': [
        'rest_framework.throttling.AnonRateThrottle',
        'rest_framework.throttling.UserRateThrottle'
    ],
    'DEFAULT_THROTTLE_RATES': {
        'anon': '10000/day',
        'user': '100000/day'
    },
    'DEFAULT_AUTHENTICATION_CLASSES': [
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ]
}

FILTERS_DEFAULT_LOOKUP_EXPR = 'icontains'


# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/
# This is defined here as a do-nothing function because we can't import
# django.utils.translation -- that module depends on the settings.
def gettext_noop(s):
    return s


LANGUAGE_CODE = 'en-us'
LANGUAGES = [
    ('af', gettext_noop('Afrikaans')),
    ('ar', gettext_noop('Arabic')),
    ('ar-dz', gettext_noop('Algerian Arabic')),
    ('ast', gettext_noop('Asturian')),
    ('az', gettext_noop('Azerbaijani')),
    ('bg', gettext_noop('Bulgarian')),
    ('be', gettext_noop('Belarusian')),
    ('bn', gettext_noop('Bengali')),
    ('br', gettext_noop('Breton')),
    ('bs', gettext_noop('Bosnian')),
    ('ca', gettext_noop('Catalan')),
    ('cs', gettext_noop('Czech')),
    ('cy', gettext_noop('Welsh')),
    ('da', gettext_noop('Danish')),
    ('de', gettext_noop('German')),
    ('dsb', gettext_noop('Lower Sorbian')),
    ('el', gettext_noop('Greek')),
    ('en', gettext_noop('English')),
    ('en-au', gettext_noop('Australian English')),
    ('en-gb', gettext_noop('British English')),
    ('eo', gettext_noop('Esperanto')),
    ('es', gettext_noop('Spanish')),
    ('es-ar', gettext_noop('Argentinian Spanish')),
    ('es-co', gettext_noop('Colombian Spanish')),
    ('es-mx', gettext_noop('Mexican Spanish')),
    ('es-ni', gettext_noop('Nicaraguan Spanish')),
    ('es-ve', gettext_noop('Venezuelan Spanish')),
    ('et', gettext_noop('Estonian')),
    ('eu', gettext_noop('Basque')),
    ('fa', gettext_noop('Persian')),
    ('fi', gettext_noop('Finnish')),
    ('fr', gettext_noop('French')),
    ('fy', gettext_noop('Frisian')),
    ('ga', gettext_noop('Irish')),
    ('gd', gettext_noop('Scottish Gaelic')),
    ('gl', gettext_noop('Galician')),
    ('he', gettext_noop('Hebrew')),
    ('hi', gettext_noop('Hindi')),
    ('hr', gettext_noop('Croatian')),
    ('hsb', gettext_noop('Upper Sorbian')),
    ('hu', gettext_noop('Hungarian')),
    ('hy', gettext_noop('Armenian')),
    ('ia', gettext_noop('Interlingua')),
    ('id', gettext_noop('Indonesian')),
    ('ig', gettext_noop('Igbo')),
    ('io', gettext_noop('Ido')),
    ('is', gettext_noop('Icelandic')),
    ('it', gettext_noop('Italian')),
    ('ja', gettext_noop('Japanese')),
    ('ka', gettext_noop('Georgian')),
    ('kab', gettext_noop('Kabyle')),
    ('kk', gettext_noop('Kazakh')),
    ('km', gettext_noop('Khmer')),
    ('kn', gettext_noop('Kannada')),
    ('ko', gettext_noop('Korean')),
    ('ky', gettext_noop('Kyrgyz')),
    ('lb', gettext_noop('Luxembourgish')),
    ('lt', gettext_noop('Lithuanian')),
    ('lv', gettext_noop('Latvian')),
    ('mk', gettext_noop('Macedonian')),
    ('ml', gettext_noop('Malayalam')),
    ('mn', gettext_noop('Mongolian')),
    ('mr', gettext_noop('Marathi')),
    ('my', gettext_noop('Burmese')),
    ('nb', gettext_noop('Norwegian Bokmål')),
    ('ne', gettext_noop('Nepali')),
    ('nl', gettext_noop('Dutch')),
    ('nn', gettext_noop('Norwegian Nynorsk')),
    ('os', gettext_noop('Ossetic')),
    ('pa', gettext_noop('Punjabi')),
    ('pl', gettext_noop('Polish')),
    ('pt', gettext_noop('Portuguese')),
    ('pt-br', gettext_noop('Brazilian Portuguese')),
    ('ro', gettext_noop('Romanian')),
    ('ru', gettext_noop('Russian')),
    ('sk', gettext_noop('Slovak')),
    ('sl', gettext_noop('Slovenian')),
    ('sq', gettext_noop('Albanian')),
    ('sr', gettext_noop('Serbian')),
    ('sr-latn', gettext_noop('Serbian Latin')),
    ('sv', gettext_noop('Swedish')),
    ('sw', gettext_noop('Swahili')),
    ('ta', gettext_noop('Tamil')),
    ('te', gettext_noop('Telugu')),
    ('tg', gettext_noop('Tajik')),
    ('th', gettext_noop('Thai')),
    ('tk', gettext_noop('Turkmen')),
    ('tr', gettext_noop('Turkish')),
    ('tt', gettext_noop('Tatar')),
    ('udm', gettext_noop('Udmurt')),
    ('uk', gettext_noop('Ukrainian')),
    ('ur', gettext_noop('Urdu')),
    ('uz', gettext_noop('Uzbek')),
    ('vi', gettext_noop('Vietnamese')),
    ('zh-hans', gettext_noop('Simplified Chinese')),
    ('zh-hant', gettext_noop('Traditional Chinese')),
]

# Languages using BiDi (right-to-left) layout
LANGUAGES_BIDI = ["he", "ar", "ar-dz", "fa", "ur"]

LOCALE_PATHS = [
    'src/geocurrency/core/locales',
    pycountry.LOCALES_DIR,
]

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = '/static/'

SENDFILE_BACKEND = 'sendfile.backends.nginx'
SENDFILE_URL = '/media'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'file': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'filename': '/var/log/geocurrencies/api-debug.log',
        },
    },
    'loggers': {
        'django': {
            'handlers': ['file'],
            'level': 'DEBUG',
            'propagate': True,
        },
    },
}

SERVICES = {
    'geocoding': {
        'pelias': 'geocurrency.countries.services.pelias.PeliasGeocoder',
        'google': 'geocurrency.countries.services.google.GoogleGeocoder',
    },
    'rates': {
        'forex': 'geocurrency.rates.services.forex.ForexService',
        'currencylayer': 'geocurrency.rates.services.currencylayer.CurrencyLayerService'
    }
}

SWAGGER_SETTINGS = {
    'SECURITY_DEFINITIONS': {
        'Token': {
            'type': 'apiKey',
            'name': 'Authorization',
            'in': 'header'
        }
    }
}

# Adapt to your local environment or override in local.py
ROOT_URLCONF = 'api.urls'
SENDFILE_ROOT = '/var/www/geocurrencies/media'
STATIC_ROOT = '/var/www/geocurrencies/static'
MEDIA_ROOT = '/var/www/geocurrencies/media'
WSGI_APPLICATION = 'api.wsgi.application'
ASGI_APPLICATION = 'api.asgi.application'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}
DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'
CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": "redis://127.0.0.1:6379/1",
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        }
    },
    "countries": {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'countries',
        'TIMEOUT': None
    }
}

from geocurrency.core.settings import *
from geocurrency.countries.settings import *
from geocurrency.currencies.settings import *
from geocurrency.rates.settings import *
from geocurrency.units.settings import *

try:
    from .local import *
except ImportError:
    pass



"""
URLs for currencies module
"""

from django.conf.urls import url, include
from rest_framework import routers

from .viewsets import CurrencyViewset

app_name = 'currencies'

router = routers.DefaultRouter()
router.register(r'', CurrencyViewset, basename='currencies')

urlpatterns = [
    url(r'^', include(router.urls)),
]



"""
Rates modules API viewsets
"""

from django.db import models
from django.db.models.functions import Extract
from django.http import HttpResponseForbidden
from django_filters import rest_framework as filters
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from geocurrency.converters.serializers import ConverterResultSerializer
from geocurrency.core.pagination import PageNumberPagination
from rest_framework import viewsets, status, mixins
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.views import APIView

from .filters import RateFilter
from .forms import RateForm
from .models import Rate, RateConverter
from .permissions import RateObjectPermission
from .serializers import RateSerializer, BulkSerializer, RateConversionPayloadSerializer, \
    RateStatSerializer


class RateViewSet(mixins.CreateModelMixin, mixins.ListModelMixin, mixins.RetrieveModelMixin,
                  viewsets.GenericViewSet):
    """
    Rate API
    """
    queryset = Rate.objects.all()
    serializer_class = RateSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = RateFilter
    pagination_class = PageNumberPagination
    permission_classes = [RateObjectPermission]
    display_page_controls = True

    def get_queryset(self):
        """
        Filter on connected user
        """
        qs = super(RateViewSet, self).get_queryset()
        if self.request.user and self.request.user.is_authenticated:
            qs = qs.filter(
                models.Q(user=self.request.user) | models.Q(user__isnull=True)
            )
        else:
            qs = qs.filter(models.Q(user__isnull=True))
        return qs

    user = openapi.Parameter(
        'user',
        openapi.IN_QUERY,
        description="Filter on user rates",
        type=openapi.TYPE_BOOLEAN)
    key = openapi.Parameter(
        'key',
        openapi.IN_QUERY,
        description="Filter on user defined category",
        type=openapi.TYPE_STRING)
    key_or_null = openapi.Parameter(
        'key_or_null',
        openapi.IN_QUERY,
        description="Show rates with defined category or no category",
        type=openapi.TYPE_STRING)
    key_isnull = openapi.Parameter(
        'key_isnull',
        openapi.IN_QUERY,
        description="ONly show results with no category",
        type=openapi.TYPE_STRING)
    value_date = openapi.Parameter(
        'value_date',
        openapi.IN_QUERY,
        description="Filter on date of the rate",
        type=openapi.TYPE_STRING)
    from_obj = openapi.Parameter(
        'from_obj',
        openapi.IN_QUERY,
        description="Filter rates with value date after date",
        type=openapi.TYPE_STRING)
    to_obj = openapi.Parameter(
        'to_obj',
        openapi.IN_QUERY,
        description="Filter rates with value date before date",
        type=openapi.TYPE_STRING)
    value = openapi.Parameter(
        'value',
        openapi.IN_QUERY,
        description="Filter rates with a specific value",
        type=openapi.TYPE_NUMBER)
    lower_bound = openapi.Parameter(
        'lower_bound',
        openapi.IN_QUERY,
        description="Filter rates with a value higher than value",
        type=openapi.TYPE_NUMBER)
    higher_bound = openapi.Parameter(
        'higher_bound',
        openapi.IN_QUERY,
        description="Filter rates with a value lower than value",
        type=openapi.TYPE_NUMBER)
    currency = openapi.Parameter(
        'currency',
        openapi.IN_QUERY,
        description="Filter by currency",
        type=openapi.TYPE_STRING)
    base_currency = openapi.Parameter(
        'base_currency',
        openapi.IN_QUERY,
        description="Filter by base currency",
        type=openapi.TYPE_STRING)

    currency_latest_values = openapi.Parameter(
        'currency_latest_values',
        openapi.IN_QUERY,
        description="Currency to filter on, limits results to latest values",
        type=openapi.TYPE_STRING)
    base_currency_latest_values = openapi.Parameter(
        'base_currency_latest_values',
        openapi.IN_QUERY,
        description="Base currency to filter on, limits results to latest values",
        type=openapi.TYPE_STRING)

    period = openapi.Parameter(
        'period',
        openapi.IN_QUERY,
        description="period to aggregate on: month, week, year, defaults to month",
        type=openapi.TYPE_STRING)

    ordering = openapi.Parameter('ordering',
                                 openapi.IN_QUERY,
                                 description="Sort on currency, base_currency, value_date, value. "
                                             "Prefix with - for descending sort",
                                 type=openapi.TYPE_STRING)

    @swagger_auto_schema(manual_parameters=[
        user, key, key_or_null, key_isnull, value_date, from_obj, to_obj, value,
        lower_bound, higher_bound, currency, base_currency,
        currency_latest_values, base_currency_latest_values, ordering],
        responses={200: RateSerializer})
    def list(self, request, *args, **kwargs):
        """
        List rates
        """
        return super().list(request, *args, **kwargs)

    @swagger_auto_schema(
        manual_parameters=[
            user, key, key_or_null, key_isnull, value_date, from_obj, to_obj, value,
            lower_bound, higher_bound, currency, base_currency,
            currency_latest_values, base_currency_latest_values, period],
        responses={200: RateStatSerializer})
    @action(['GET'], detail=False, url_path='stats', url_name='stats')
    def stats(self, request, *args, **kwargs):
        """
        stats on rates
        """
        period = request.GET.get('period', 'month')
        if period not in ['week', 'month', 'year']:
            return Response("Invalid period", status=status.HTTP_400_BAD_REQUEST)
        rate_filter = RateFilter(request.GET, queryset=self.queryset, request=request)
        qs = rate_filter.qs.values('currency', 'base_currency')
        if period == 'month':
            qs = qs.annotate(month=Extract('value_date', 'month'))
        if period == 'week':
            qs = qs.annotate(week=Extract('value_date', 'week'))
        qs = qs.annotate(
            year=Extract('value_date', 'year'),
            avg=models.Avg('value'),
            max=models.Max('value'),
            min=models.Min('value'),
            std_dev=models.StdDev('value')
        ).order_by('-year', '-' + period)
        results = [
            {
                'currency': result['currency'],
                'base_currency': result['base_currency'],
                'period': f"{result['year']}-{str(result[period]).zfill(2)}"
                if period != 'year'
                else result['year'],
                'avg': result['avg'],
                'max': result['max'],
                'min': result['min'],
                'std_dev': result['std_dev'],
            }
            for result in qs
        ]
        data = {
            'key': request.GET.get('key'),
            'period': period,
            'from_date': request.GET.get('from_date', ''),
            'to_date': request.GET.get('to_date', ''),
            'results': results
        }
        serializer = RateStatSerializer(data)
        return Response(serializer.data, content_type="application/json")

    def create(self, request, *args, **kwargs):
        """
        Create a new rate
        """
        rate_form = RateForm(request.data)
        if rate_form.is_valid():
            rate = rate_form.save(commit=False)
            if request.user and request.user.is_authenticated:
                rate.user = request.user
                rate.save()
                serializer = RateSerializer(rate)
                return Response(serializer.data, status=status.HTTP_201_CREATED)
            else:
                return HttpResponseForbidden()
        else:
            return Response(rate_form.errors, status=status.HTTP_400_BAD_REQUEST,
                            content_type="application/json")

    @swagger_auto_schema(method='post', request_body=BulkSerializer,
                         responses={201: RateSerializer})
    @action(['POST'], detail=False, url_path='bulk', url_name="bulk_create")
    def create_bulk(self, request):
        """
        Create rates for user on a range of dates
        """
        if not request.user or not request.user.is_authenticated:
            return HttpResponseForbidden()
        bs = BulkSerializer(data=request.data)
        if not bs.is_valid():
            return Response(bs.errors, status=status.HTTP_400_BAD_REQUEST)
        bulk_rate = bs.create(validated_data=bs.validated_data)
        rates = bulk_rate.to_rates(user=request.user)
        serializer = RateSerializer(rates, many=True)
        return Response(serializer.data, content_type="application/json",
                        status=status.HTTP_201_CREATED)


class ConvertView(APIView):
    """
    Conversion API
    """

    @swagger_auto_schema(request_body=RateConversionPayloadSerializer,
                         responses={200: ConverterResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency and date to a reference currency
        :param request: HTTP request
        """
        cps = RateConversionPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST,
                            content_type="application/json")
        cp = cps.create(cps.validated_data)
        try:
            converter = RateConverter.load(cp.batch_id)
        except KeyError:
            converter = RateConverter(
                id=cp.batch_id,
                user=request.user,
                key=cp.key,
                base_currency=cp.target
            )
        if cp.data:
            errors = converter.add_data(data=cp.data)
            if errors:
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = converter.convert()
            serializer = ConverterResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': converter.id, 'status': converter.status},
                            content_type="application/json")



"""
Standard classes for the Converter module
"""

import logging
import pickle
import uuid

from django.core.cache import cache


class ConverterLoadError(Exception):
    """
    Exception when loading a converter from its redis pickle
    """
    msg = 'Error while loading converter'


class BaseConverter:
    """
    Base class for conversion
    Mock up for usage in type hinting
    """
    INITIATED_STATUS = 'initiated'
    INSERTING_STATUS = 'inserting'
    PENDING_STATUS = 'pending'
    FINISHED = 'finished'
    WITH_ERRORS = 'finished with errors'


class ConverterResultDetail:
    """
    Details of a conversion
    """
    unit = None
    original_value = 0
    date = None
    conversion_rate = 0
    converted_value = 0

    def __init__(self, unit: str, original_value: float, date: date, conversion_rate: float,
                 converted_value: float):
        """
        Initialize details
        :param unit: dimension as a string
        :param original_value: value before conversion
        :param date: date of conversion
        :param conversion_rate: rate of conversion
        :param converted_value: resulting value
        """
        self.unit = unit
        self.original_value = original_value
        self.date = date
        self.conversion_rate = conversion_rate
        self.converted_value = converted_value


class ConverterResultError:
    """
    Error from a conversion
    """
    unit = None
    original_value = None
    date = None
    error = None

    def __init__(self, unit: str, original_value: float, date: date, error: str):
        """
        Initialize error
        :param unit: string of the dimension
        :param original_value: value before conversion
        :param date: date of conversion
        :param error: description of the error
        """
        self.unit = unit
        self.original_value = original_value
        self.date = date
        self.error = error


class ConverterResult:
    """
    Result of a batch of conversions
    """
    id = None
    target = None
    detail = []
    sum = 0
    status = None
    errors = []

    def __init__(self, id: str = None, target: str = None, detail: [ConverterResultDetail] = None,
                 sum: float = 0, status: str = BaseConverter.INITIATED_STATUS,
                 errors: [ConverterResultError] = None):
        """
        Initialize result
        :param id: ID of the batch
        :param target: target currency
        :param detail: List of ConverterResultDetail
        :param sum: sum of all detailed conversions
        :param status: status of the batch
        :param errors: List of conversion errors
        """
        self.id = id
        self.target = target
        self.detail = detail or []
        self.sum = sum
        self.status = status
        self.errors = errors or []

    def increment_sum(self, value):
        """
        Sum individual conversion results
        They are all in the target currency
        :param value: result of a conversion
        """
        try:
            float(value)
            self.sum += value
        except ValueError:
            logging.error("invalid value, will not increment result sum", value)

    def end_batch(self):
        """
        Puts a final status on the batch
        """
        if self.errors:
            self.status = BaseConverter.WITH_ERRORS
        else:
            self.status = BaseConverter.FINISHED
        return self.status


class BaseConverter:
    """
    Base conversion class
    """
    INITIATED_STATUS = 'initiated'
    INSERTING_STATUS = 'inserting'
    PENDING_STATUS = 'pending'
    FINISHED = 'finished'
    WITH_ERRORS = 'finished with errors'
    id = None
    status = INITIATED_STATUS
    data = []
    converted_lines = []
    aggregated_result = {}

    def __init__(self, id: str = None):
        """
        Initialize BaseConverter
        :param id: ID of the batch
        """
        self.id = id or uuid.uuid4()
        self.data = []

    @classmethod
    def load(cls, id: str) -> BaseConverter:
        """
        Load Converter from cache
        :param id: ID of the batch
        """
        obj = cache.get(id)
        if obj:
            return pickle.loads(obj)
        raise KeyError(f"Converter with id {id} not found in cache")

    def save(self):
        """
        Save Converter to cache
        """
        cache.set(self.id, pickle.dumps(self))

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        :param data: list of items to convert
        """
        if not data:
            return [{'data': 'Empty data set', }]
        errors = self.check_data(data)
        if errors:
            return errors
        self.status = self.INSERTING_STATUS
        self.save()
        return []

    def end_batch(self, status: str):
        """
        set status of the batch
        :param status: status from statuses
        """
        self.status = status

    def check_data(self, data):
        """
        Validates data
        Not implementd
        :param data: list of items to convert
        """
        raise NotImplementedError

    def convert(self) -> ConverterResult:
        """
        Converts data to base currency
        Not implemented
        """
        raise NotImplementedError


class Batch:
    """
    Batch class
    """
    id = None
    status = None

    def __init__(self, id: str, status: str):
        """
        Initialize the batch
        :param id: ID of the batch
        :param status: status of the batch
        """
        self.id = id
        self.status = status



"""GeoCurrencies URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf.urls import url, include
from django.urls import path
from .converters.views import WatchView
from .countries import urls as country_urls
from .currencies import urls as currency_urls
from .rates import urls as rate_urls
from .units import urls as unit_urls

urlpatterns = [
    path('currencies/', include(currency_urls)),
    path('countries/', include(country_urls)),
    path('rates/', include(rate_urls)),
    path('units/', include(unit_urls)),
    url(r'^watch/(?P<converter_id>[0-9a-f-]{36})/$', WatchView.as_view()),
]



"""
Renderer for Units
"""

import json
from rest_framework.renderers import JSONRenderer
from rest_framework.settings import api_settings
from rest_framework.utils.encoders import JSONEncoder

from .models import Operand


class UnitsEncoder(JSONEncoder):
    """
    Encoder for Custom Unit
    """

    def default(self, obj):
        if isinstance(obj, Operand):
            return json.dumps({'name': obj.name, 'value': obj.value, 'unit': obj.unit})
        else:
            return super().default(obj)


class UnitsRenderer(JSONRenderer):
    """
    Renderer which serializes to JSON with UnitsEncoder
    """
    media_type = 'application/json'
    format = 'json'
    encoder_class = UnitsEncoder
    ensure_ascii = not api_settings.UNICODE_JSON
    compact = api_settings.COMPACT_JSON
    strict = api_settings.STRICT_JSON

    # We don't set a charset because JSON is a binary encoding,
    # that can be encoded as utf-8, utf-16 or utf-32.
    # See: https://www.ietf.org/rfc/rfc4627.txt
    # Also: http://lucumr.pocoo.org/2013/7/19/application-mimetypes-and-encodings/
    charset = None
